# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lxdbapi/responses.proto
"""Generated protocol buffer code."""
from ._vendor.google.protobuf import descriptor as _descriptor
from ._vendor.google.protobuf import descriptor_pool as _descriptor_pool
from ._vendor.google.protobuf import message as _message
from ._vendor.google.protobuf import reflection as _reflection
from ._vendor.google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from lxdbapi import common_pb2 as lxdbapi_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17lxdbapi/responses.proto\x12\x07lxdbapi\x1a\x14lxdbapi/common.proto\"\xe1\x01\n\x11ResultSetResponse\x12\x15\n\rconnection_id\x18\x01 \x01(\t\x12\x14\n\x0cstatement_id\x18\x02 \x01(\r\x12\x15\n\rown_statement\x18\x03 \x01(\x08\x12%\n\tsignature\x18\x04 \x01(\x0b\x32\x12.lxdbapi.Signature\x12#\n\x0b\x66irst_frame\x18\x05 \x01(\x0b\x32\x0e.lxdbapi.Frame\x12\x14\n\x0cupdate_count\x18\x06 \x01(\x04\x12&\n\x08metadata\x18\x07 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"\x81\x01\n\x0f\x45xecuteResponse\x12+\n\x07results\x18\x01 \x03(\x0b\x32\x1a.lxdbapi.ResultSetResponse\x12\x19\n\x11missing_statement\x18\x02 \x01(\x08\x12&\n\x08metadata\x18\x03 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"f\n\x0fPrepareResponse\x12+\n\tstatement\x18\x01 \x01(\x0b\x32\x18.lxdbapi.StatementHandle\x12&\n\x08metadata\x18\x02 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"\x8a\x01\n\rFetchResponse\x12\x1d\n\x05\x66rame\x18\x01 \x01(\x0b\x32\x0e.lxdbapi.Frame\x12\x19\n\x11missing_statement\x18\x02 \x01(\x08\x12\x17\n\x0fmissing_results\x18\x03 \x01(\x08\x12&\n\x08metadata\x18\x04 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"n\n\x17\x43reateStatementResponse\x12\x15\n\rconnection_id\x18\x01 \x01(\t\x12\x14\n\x0cstatement_id\x18\x02 \x01(\r\x12&\n\x08metadata\x18\x03 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"@\n\x16\x43loseStatementResponse\x12&\n\x08metadata\x18\x01 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"@\n\x16OpenConnectionResponse\x12&\n\x08metadata\x18\x01 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"A\n\x17\x43loseConnectionResponse\x12&\n\x08metadata\x18\x01 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"s\n\x16\x43onnectionSyncResponse\x12\x31\n\nconn_props\x18\x01 \x01(\x0b\x32\x1d.lxdbapi.ConnectionProperties\x12&\n\x08metadata\x18\x02 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"\x8d\x01\n\x17\x44\x61tabasePropertyElement\x12&\n\x03key\x18\x01 \x01(\x0b\x32\x19.lxdbapi.DatabaseProperty\x12\"\n\x05value\x18\x02 \x01(\x0b\x32\x13.lxdbapi.TypedValue\x12&\n\x08metadata\x18\x03 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"s\n\x18\x44\x61tabasePropertyResponse\x12/\n\x05props\x18\x01 \x03(\x0b\x32 .lxdbapi.DatabasePropertyElement\x12&\n\x08metadata\x18\x02 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"\xc6\x01\n\rErrorResponse\x12\x12\n\nexceptions\x18\x01 \x03(\t\x12\x16\n\x0ehas_exceptions\x18\x07 \x01(\x08\x12\x15\n\rerror_message\x18\x02 \x01(\t\x12#\n\x08severity\x18\x03 \x01(\x0e\x32\x11.lxdbapi.Severity\x12\x12\n\nerror_code\x18\x04 \x01(\r\x12\x11\n\tsql_state\x18\x05 \x01(\t\x12&\n\x08metadata\x18\x06 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"n\n\x13SyncResultsResponse\x12\x19\n\x11missing_statement\x18\x01 \x01(\x08\x12\x14\n\x0cmore_results\x18\x02 \x01(\x08\x12&\n\x08metadata\x18\x03 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"%\n\x0bRpcMetadata\x12\x16\n\x0eserver_address\x18\x01 \x01(\t\"\x10\n\x0e\x43ommitResponse\"\x12\n\x10RollbackResponse\"\x9d\x01\n\x14\x45xecuteBatchResponse\x12\x15\n\rconnection_id\x18\x01 \x01(\t\x12\x14\n\x0cstatement_id\x18\x02 \x01(\r\x12\x15\n\rupdate_counts\x18\x03 \x03(\x04\x12\x19\n\x11missing_statement\x18\x04 \x01(\x08\x12&\n\x08metadata\x18\x05 \x01(\x0b\x32\x14.lxdbapi.RpcMetadata\"9\n\x10LxActionResponse\x12%\n\x06params\x18\x01 \x03(\x0b\x32\x15.lxdbapi.KeyValuePairB\"\n org.apache.calcite.avatica.protob\x06proto3')



_RESULTSETRESPONSE = DESCRIPTOR.message_types_by_name['ResultSetResponse']
_EXECUTERESPONSE = DESCRIPTOR.message_types_by_name['ExecuteResponse']
_PREPARERESPONSE = DESCRIPTOR.message_types_by_name['PrepareResponse']
_FETCHRESPONSE = DESCRIPTOR.message_types_by_name['FetchResponse']
_CREATESTATEMENTRESPONSE = DESCRIPTOR.message_types_by_name['CreateStatementResponse']
_CLOSESTATEMENTRESPONSE = DESCRIPTOR.message_types_by_name['CloseStatementResponse']
_OPENCONNECTIONRESPONSE = DESCRIPTOR.message_types_by_name['OpenConnectionResponse']
_CLOSECONNECTIONRESPONSE = DESCRIPTOR.message_types_by_name['CloseConnectionResponse']
_CONNECTIONSYNCRESPONSE = DESCRIPTOR.message_types_by_name['ConnectionSyncResponse']
_DATABASEPROPERTYELEMENT = DESCRIPTOR.message_types_by_name['DatabasePropertyElement']
_DATABASEPROPERTYRESPONSE = DESCRIPTOR.message_types_by_name['DatabasePropertyResponse']
_ERRORRESPONSE = DESCRIPTOR.message_types_by_name['ErrorResponse']
_SYNCRESULTSRESPONSE = DESCRIPTOR.message_types_by_name['SyncResultsResponse']
_RPCMETADATA = DESCRIPTOR.message_types_by_name['RpcMetadata']
_COMMITRESPONSE = DESCRIPTOR.message_types_by_name['CommitResponse']
_ROLLBACKRESPONSE = DESCRIPTOR.message_types_by_name['RollbackResponse']
_EXECUTEBATCHRESPONSE = DESCRIPTOR.message_types_by_name['ExecuteBatchResponse']
_LXACTIONRESPONSE = DESCRIPTOR.message_types_by_name['LxActionResponse']
ResultSetResponse = _reflection.GeneratedProtocolMessageType('ResultSetResponse', (_message.Message,), {
  'DESCRIPTOR' : _RESULTSETRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.ResultSetResponse)
  })
_sym_db.RegisterMessage(ResultSetResponse)

ExecuteResponse = _reflection.GeneratedProtocolMessageType('ExecuteResponse', (_message.Message,), {
  'DESCRIPTOR' : _EXECUTERESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.ExecuteResponse)
  })
_sym_db.RegisterMessage(ExecuteResponse)

PrepareResponse = _reflection.GeneratedProtocolMessageType('PrepareResponse', (_message.Message,), {
  'DESCRIPTOR' : _PREPARERESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.PrepareResponse)
  })
_sym_db.RegisterMessage(PrepareResponse)

FetchResponse = _reflection.GeneratedProtocolMessageType('FetchResponse', (_message.Message,), {
  'DESCRIPTOR' : _FETCHRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.FetchResponse)
  })
_sym_db.RegisterMessage(FetchResponse)

CreateStatementResponse = _reflection.GeneratedProtocolMessageType('CreateStatementResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATESTATEMENTRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.CreateStatementResponse)
  })
_sym_db.RegisterMessage(CreateStatementResponse)

CloseStatementResponse = _reflection.GeneratedProtocolMessageType('CloseStatementResponse', (_message.Message,), {
  'DESCRIPTOR' : _CLOSESTATEMENTRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.CloseStatementResponse)
  })
_sym_db.RegisterMessage(CloseStatementResponse)

OpenConnectionResponse = _reflection.GeneratedProtocolMessageType('OpenConnectionResponse', (_message.Message,), {
  'DESCRIPTOR' : _OPENCONNECTIONRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.OpenConnectionResponse)
  })
_sym_db.RegisterMessage(OpenConnectionResponse)

CloseConnectionResponse = _reflection.GeneratedProtocolMessageType('CloseConnectionResponse', (_message.Message,), {
  'DESCRIPTOR' : _CLOSECONNECTIONRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.CloseConnectionResponse)
  })
_sym_db.RegisterMessage(CloseConnectionResponse)

ConnectionSyncResponse = _reflection.GeneratedProtocolMessageType('ConnectionSyncResponse', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTIONSYNCRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.ConnectionSyncResponse)
  })
_sym_db.RegisterMessage(ConnectionSyncResponse)

DatabasePropertyElement = _reflection.GeneratedProtocolMessageType('DatabasePropertyElement', (_message.Message,), {
  'DESCRIPTOR' : _DATABASEPROPERTYELEMENT,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.DatabasePropertyElement)
  })
_sym_db.RegisterMessage(DatabasePropertyElement)

DatabasePropertyResponse = _reflection.GeneratedProtocolMessageType('DatabasePropertyResponse', (_message.Message,), {
  'DESCRIPTOR' : _DATABASEPROPERTYRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.DatabasePropertyResponse)
  })
_sym_db.RegisterMessage(DatabasePropertyResponse)

ErrorResponse = _reflection.GeneratedProtocolMessageType('ErrorResponse', (_message.Message,), {
  'DESCRIPTOR' : _ERRORRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.ErrorResponse)
  })
_sym_db.RegisterMessage(ErrorResponse)

SyncResultsResponse = _reflection.GeneratedProtocolMessageType('SyncResultsResponse', (_message.Message,), {
  'DESCRIPTOR' : _SYNCRESULTSRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.SyncResultsResponse)
  })
_sym_db.RegisterMessage(SyncResultsResponse)

RpcMetadata = _reflection.GeneratedProtocolMessageType('RpcMetadata', (_message.Message,), {
  'DESCRIPTOR' : _RPCMETADATA,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.RpcMetadata)
  })
_sym_db.RegisterMessage(RpcMetadata)

CommitResponse = _reflection.GeneratedProtocolMessageType('CommitResponse', (_message.Message,), {
  'DESCRIPTOR' : _COMMITRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.CommitResponse)
  })
_sym_db.RegisterMessage(CommitResponse)

RollbackResponse = _reflection.GeneratedProtocolMessageType('RollbackResponse', (_message.Message,), {
  'DESCRIPTOR' : _ROLLBACKRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.RollbackResponse)
  })
_sym_db.RegisterMessage(RollbackResponse)

ExecuteBatchResponse = _reflection.GeneratedProtocolMessageType('ExecuteBatchResponse', (_message.Message,), {
  'DESCRIPTOR' : _EXECUTEBATCHRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.ExecuteBatchResponse)
  })
_sym_db.RegisterMessage(ExecuteBatchResponse)

LxActionResponse = _reflection.GeneratedProtocolMessageType('LxActionResponse', (_message.Message,), {
  'DESCRIPTOR' : _LXACTIONRESPONSE,
  '__module__' : 'lxdbapi.responses_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.LxActionResponse)
  })
_sym_db.RegisterMessage(LxActionResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n org.apache.calcite.avatica.proto'
  _RESULTSETRESPONSE._serialized_start=59
  _RESULTSETRESPONSE._serialized_end=284
  _EXECUTERESPONSE._serialized_start=287
  _EXECUTERESPONSE._serialized_end=416
  _PREPARERESPONSE._serialized_start=418
  _PREPARERESPONSE._serialized_end=520
  _FETCHRESPONSE._serialized_start=523
  _FETCHRESPONSE._serialized_end=661
  _CREATESTATEMENTRESPONSE._serialized_start=663
  _CREATESTATEMENTRESPONSE._serialized_end=773
  _CLOSESTATEMENTRESPONSE._serialized_start=775
  _CLOSESTATEMENTRESPONSE._serialized_end=839
  _OPENCONNECTIONRESPONSE._serialized_start=841
  _OPENCONNECTIONRESPONSE._serialized_end=905
  _CLOSECONNECTIONRESPONSE._serialized_start=907
  _CLOSECONNECTIONRESPONSE._serialized_end=972
  _CONNECTIONSYNCRESPONSE._serialized_start=974
  _CONNECTIONSYNCRESPONSE._serialized_end=1089
  _DATABASEPROPERTYELEMENT._serialized_start=1092
  _DATABASEPROPERTYELEMENT._serialized_end=1233
  _DATABASEPROPERTYRESPONSE._serialized_start=1235
  _DATABASEPROPERTYRESPONSE._serialized_end=1350
  _ERRORRESPONSE._serialized_start=1353
  _ERRORRESPONSE._serialized_end=1551
  _SYNCRESULTSRESPONSE._serialized_start=1553
  _SYNCRESULTSRESPONSE._serialized_end=1663
  _RPCMETADATA._serialized_start=1665
  _RPCMETADATA._serialized_end=1702
  _COMMITRESPONSE._serialized_start=1704
  _COMMITRESPONSE._serialized_end=1720
  _ROLLBACKRESPONSE._serialized_start=1722
  _ROLLBACKRESPONSE._serialized_end=1740
  _EXECUTEBATCHRESPONSE._serialized_start=1743
  _EXECUTEBATCHRESPONSE._serialized_end=1900
  _LXACTIONRESPONSE._serialized_start=1902
  _LXACTIONRESPONSE._serialized_end=1959
# @@protoc_insertion_point(module_scope)



import logging
import random
import re
import time
import uuid
from threading import RLock

from lxdbapi.balanced_connection_manager import BalancedConnectionManager
from lxdbapi.connection_manager import ConnectionManager, MultiConnectionI, ServiceConnectionI

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from lxdbapi.client import AvaticaClient
from lxdbapi.connection import Connection

__all__ = ['connection_manager_info', 'connection_manager_stop', 'MultiConnection', 'MultiCursor']

logger = logging.getLogger(__name__)


class ConnectionManagerHolder(ConnectionManager):
    _delegator = None

    def is_inited(self):
        return self._delegator is not None

    def init_url(self, url, info: dict):
        if url and self._delegator is None:
            service_info = info.copy()
            service_info.pop('distribute', None)
            conn = ServiceConnection(url, service_info)
            self._delegator = BalancedConnectionManager(conn)
            self._delegator.add_connection(conn)  # Adding service connection to CM. Maybe we shouldn't

    def add_connection(self, conn):
        if self._delegator:
            self._delegator.add_connection(conn)

    def remove_connection(self, conn):
        if self._delegator:
            self._delegator.remove_connection(conn)

    def reconnect(self, conn):
        if self._delegator:
            return self._delegator.reconnect(conn)
        else:
            raise RuntimeError("No Connection manager")

    def update(self, addresses: set):
        if self._delegator:
            self._delegator.update(addresses)

    def next_qe(self):
        if self._delegator:
            return self._delegator.next_qe()
        return None

    def invalidate_address(self, addr: str):
        if self._delegator:
            self._delegator.invalidate_address(addr)

    def validate_address(self, addr: str):
        if self._delegator:
            self._delegator.validate_address(addr)

    def is_valid_address(self, addr: str):
        if self._delegator:
            return self._delegator.is_valid_address(addr)
        return True

    def is_stopped(self):
        if self._delegator:
            return self._delegator.is_stopped()
        return False

    def stop(self):
        if self._delegator:
            self._delegator.stop()

    def info(self):
        if self._delegator:
            return self._delegator.info()
        return "No connection manager"


_connection_manager = ConnectionManagerHolder()


def connection_manager(url=None, info=None):
    if url:
        _connection_manager.init_url(url, info)
    return _connection_manager


def connection_manager_info():
    return connection_manager().info()


def connection_manager_stop():
    return connection_manager().stop()


def _invoke_with_reconnect(multi_conn, cursor, func):
    manager = connection_manager(None)
    if multi_conn.is_do_reconnect():
        is_reconnected = manager.reconnect(multi_conn)
        if is_reconnected:
            if cursor:
                cursor.reset()
        else:
            logger.debug('Not reconnected but retry on current one: {}'.format(multi_conn.qe_address()))
    retry = 4
    result = None
    while retry > 0:
        retry -= 1
        try:
            result = func()
            manager.validate_address(multi_conn.qe_address())
            break
        except BaseException as ex:
            msg = str(ex)
            if 'IO Error' in msg or 'Connection refused' in msg \
                    or 'Connection reset by peer' in msg or 'Cannot create cursor' in msg:
                logger.debug('reconnect cause of: {}'.format(msg))
                if manager.is_inited() and not manager.is_stopped():
                    manager.invalidate_address(multi_conn.qe_address())
                    is_reconnected = manager.reconnect(multi_conn)
                    if is_reconnected:
                        if cursor:
                            cursor.reset()
                    else:
                        raise RuntimeError("Cannot reconnect {}: {}".format(multi_conn.id(), ex), ex)
                else:
                    multi_conn.reconnect()
                    if cursor:
                        cursor.reset()
            elif 'the cursor is already closed' in msg and not cursor.closed:
                logger.debug('Delegator closed so retry it again: {}'.format(ex))
                cursor.reset()
            elif 'Connection already exists' in msg \
                    or 'Error parsing message' in msg or 'Bad file descriptor' in msg \
                    or 'RPC request failed on ParseResult' in msg \
                    or '\'NoneType\' object has no attribute \'makefile\'' in msg \
                    or '\'NoneType\' object has no attribute \'setsockopt\'' in msg \
                    or 'unexpected response type ' \
                       '"org.apache.calcite.avatica.proto.Responses$CloseStatementResponse"' in msg:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception(ex)
                logger.debug('Ignore exception and retry again: {}'.format(ex))
            else:
                raise RuntimeError("Execution failed: {}".format(ex), ex)
    return result


def change_url_address(url, addr):
    url_parsed = urlparse.urlparse(url)
    if addr == url_parsed.netloc:
        return url
    replaced = url_parsed._replace(netloc=addr)
    return urlparse.urlunparse(list(replaced))


class MultiConnection(MultiConnectionI):

    def __init__(self, url, info):
        self._delegator = None
        self._cursors = []
        self._qe_address = None
        self._do_reconnect_lock = RLock()
        self._is_abort_pending_txn = False
        self._is_pending_commit = False
        self._is_auto_commit = True
        self._do_reconnect = False
        self._force_reopen = False
        self._closed = False
        self._info = info
        self._url = url
        m = re.search(r'leanxcale://(.*)@', url)
        if m is not None:
            user = m.group(1)
            userpwdmatch = re.search(r'(.*):(.*)', user)
            if userpwdmatch is not None:
                userpwd = user.split(":")
                self._info['user'] = userpwdmatch.group(1)
                self._info['password'] = userpwdmatch.group(2)
            else:
                self._info['user'] = user
        self._id = str(uuid.uuid4())
        self._open()

    def _connect(self):
        logger.debug("Connecting to address {}. Url = {}".format(self._qe_address, str(self._url)))
        client = AvaticaClient(self._url, 5, None, **self._info)
        self._delegator = Connection(client, self.id(), **self._info)
        self._delegator.open()
        logger.debug("Connected to {}".format(client.url))
        for cursor in self._cursors:
            if cursor is not None and not cursor.closed:
                cursor.reset()

    def _open(self):  # call just once
        distribute = self._info.get('distribute', None)
        if distribute:
            distribute_lower = str(distribute).lower()
            if distribute_lower == 'yes' or distribute_lower == 'true' or distribute_lower == 'balanced':
                addr = connection_manager(self.url(), self._info).next_qe()
                if addr is None or addr == "":
                    self._qe_address = urlparse.urlparse(self.url()).netloc
                else:
                    self._url = change_url_address(self.url(), addr)
                    self._qe_address = addr
            else:
                self._qe_address = urlparse.urlparse(self.url()).netloc

        else:
            self._qe_address = urlparse.urlparse(self.url()).netloc
        logger.debug("connecting to {}".format(self._qe_address))
        self._connect()
        with self._do_reconnect_lock:
            self._is_auto_commit = self._delegator.autocommit
        connection_manager().add_connection(self)

    def __del__(self):
        if self._delegator and not self._closed:
            try:
                self.close()
            except BaseException as ex:
                logger.debug("Ignoring exception while closing: {}".format(ex))
            finally:
                self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._delegator and not self._closed:
            self.close()

    def id(self):
        return self._id

    def url(self):
        return self._url

    def qe_address(self):
        return self._qe_address

    def info(self):
        return self._info

    def close(self):
        if not self._closed:
            self._closed = True
            connection_manager().remove_connection(self)
            try:
                for cursor in self._cursors:
                    if cursor is not None and not cursor.closed:
                        try:
                            cursor.close()
                        except BaseException as ex:
                            logger.debug('Ignoring exception and forcing cursor reset: {}'.format(ex))
                            cursor.reset()
            finally:
                self._delegator.close()

    @property
    def closed(self):
        return self._closed

    def commit(self):
        is_pending_changes = False
        with self._do_reconnect_lock:
            self._is_pending_commit = False
            if self._is_abort_pending_txn:
                is_pending_changes = True
            self._is_abort_pending_txn = False

        if is_pending_changes:
            raise RuntimeError("Pending changes when reconnected to " + self.qe_address())

        _invoke_with_reconnect(self, None, lambda: self._delegator.commit())

    def rollback(self):
        is_pending_changes = False
        with self._do_reconnect_lock:
            if self._is_abort_pending_txn:  # no need to abort pending txn
                self._is_abort_pending_txn = False
                self._is_pending_commit = False
                is_pending_changes = True

        if is_pending_changes is False:
            _invoke_with_reconnect(self, None, lambda: self._delegator.rollback())
        with self._do_reconnect_lock:
            self._is_pending_commit = False

    def cursor(self, cursor_factory=None):
        result = MultiCursor(self, cursor_factory)
        self._cursors.append(result)
        return result

    def set_session(self, autocommit=None, readonly=None, useTZ=False):
        _invoke_with_reconnect(self, None, lambda: self._delegator.set_session(autocommit, readonly, useTZ))
        if autocommit is not None:
            with self._do_reconnect_lock:
                self._is_auto_commit = self._delegator.autocommit

    @property
    def autocommit(self):
        with self._do_reconnect_lock:
            result = self._is_auto_commit
        return result

    @autocommit.setter
    def autocommit(self, value):
        if self.autocommit is not value:
            _invoke_with_reconnect(self, None, lambda: self._delegator.autocommit(value))
            with self._do_reconnect_lock:
                self._is_auto_commit = self._delegator.autocommit

    @property
    def readonly(self):
        return self.readonly

    @readonly.setter
    def readonly(self, value):
        _invoke_with_reconnect(self, None, lambda: self._delegator.readonly(value))

    @property
    def transactionisolation(self):
        return self._delegator.transactionisolation

    @property
    def useTZ(self):
        return self._delegator.useTZ

    def reconnect(self, addr=None):
        already_connected = False
        if addr is not None:
            with self._do_reconnect_lock:
                if self._force_reopen is False and self._qe_address == addr:
                    logger.debug("Already connected to " + str(self._qe_address))
                    already_connected = True
                else:
                    if self._is_auto_commit is False and self._is_pending_commit:
                        self._is_abort_pending_txn = True
                self._is_pending_commit = False
        if already_connected is False:
            logger.debug("reconnecting to ({} or {})".format(addr, self._qe_address))
            if addr:
                self._url = change_url_address(self._url, addr)
                self._qe_address = addr
            self._connect()
            is_auto_commit = self.autocommit
            if is_auto_commit != self._delegator.autocommit:
                self._delegator.set_session(autocommit=is_auto_commit)
        connection_manager().add_connection(self)

    def is_abort_pending_txn(self):
        with self._do_reconnect_lock:
            result = self._is_abort_pending_txn
        return result

    def is_pending_commit(self):
        with self._do_reconnect_lock:
            result = self._is_pending_commit
        return result

    def is_do_reconnect(self):
        with self._do_reconnect_lock:
            result = self._do_reconnect and self._is_pending_commit is False
        return result

    def is_force_reopen(self):
        with self._do_reconnect_lock:
            result = self._force_reopen
        return result

    def txn_update(self):
        with self._do_reconnect_lock:
            if self._is_auto_commit is False and self._is_pending_commit is False:
                self._is_pending_commit = True

    def do_reconnect(self):
        with self._do_reconnect_lock:
            self._do_reconnect = True
            self._force_reopen = False

    def invalidate_address(self):
        with self._do_reconnect_lock:
            self._do_reconnect = True
            self._force_reopen = self._is_pending_commit is False

    def unset_do_reconnect(self):
        with self._do_reconnect_lock:
            self._do_reconnect = False
            self._force_reopen = False

    @property
    def delegator(self):
        return self._delegator

    def connection_manager_info(self):
        return connection_manager().info()


class MultiCursor(object):

    def __init__(self, conn: MultiConnection, cursor_factory):
        self._closed = False
        self._connection = conn
        self._id = random.randint(0, 9999999)
        self._cursor_factory = cursor_factory
        self._delegator = None

    def __del__(self):
        if not self._connection.closed and not self._closed:
            try:
                self.close()
            except BaseException as ex:
                logger.debug("Ignoring exception while closing: {}".format(ex))
            finally:
                self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            if not self.closed:
                self.close()
        finally:
            self._closed = True

    def __iter__(self):
        return self

    def __next__(self):
        row = self.fetchone()
        if row is None:
            raise StopIteration
        return row

    next = __next__

    def get_connection(self):
        return self._connection

    def _get_delegator(self):
        if self._delegator is not None and self._delegator.closed and not self._closed:
            logger.debug("recreate cursor {}".format(self._id))
            self._delegator = None
        retries = 5
        while self._delegator is None and retries > 0:
            retries -= 1
            if self._connection.closed:
                raise RuntimeError('Connection is closed. No cursor available')
            try:
                self._delegator = self._connection.delegator.cursor(self._cursor_factory)
                self._delegator.set_xid(self._id)
            except BaseException as ex:
                logger.warning("Couldn't create cursor. Retrying.\n {}".format(ex))
                time.sleep(2 / 10)  # TODO make it configurable

        if self._delegator is None:
            raise RuntimeError('Cannot create cursor')
        return self._delegator

    def reset(self):
        logger.debug("reset cursor {}".format(self._id))
        self._delegator = None

    def close(self):
        try:
            _invoke_with_reconnect(self.get_connection(), self, lambda: self._get_delegator().close())
        finally:
            self._closed = True

    @property
    def closed(self):
        return self._closed

    @property
    def description(self):
        return self._get_delegator().description

    def get_tables(self, catalog=None, schemaPattern=None, tableNamePattern=None, typeList=None):
        return _invoke_with_reconnect(self.get_connection(), self,
                                             lambda: self._get_delegator().get_tables(catalog, schemaPattern,
                                                                                      tableNamePattern,
                                                                                      typeList))

    def get_schemas(self, catalog=None, schemaPattern=None):
        return _invoke_with_reconnect(self.get_connection(), self,
                                             lambda: self._get_delegator().get_schemas(catalog, schemaPattern))

    def get_catalogs(self):
        return _invoke_with_reconnect(self.get_connection(), self, lambda: self._get_delegator().get_catalogs())

    def get_columns(self, catalog=None, schemaPattern=None, tableNamePattern=None, columnNamePattern=None):
        return _invoke_with_reconnect(self.get_connection(), self,
                                             lambda: self._get_delegator().get_columns(catalog, schemaPattern,
                                                                                       tableNamePattern,
                                                                                       columnNamePattern))

    def get_table_types(self):
        return _invoke_with_reconnect(self.get_connection(), self, lambda: self._get_delegator().get_table_types())

    def get_type_info(self):
        return _invoke_with_reconnect(self.get_connection(), self, lambda: self._get_delegator().get_type_info())

    def execute(self, operation, parameters=None):
        result = _invoke_with_reconnect(self.get_connection(), self, lambda: self._get_delegator().execute(operation, parameters))
        self._connection.txn_update()
        return result

    def executemany(self, operation, seq_of_parameters):
        result = _invoke_with_reconnect(self.get_connection(), self,
                                               lambda: self._get_delegator().executemany(operation, seq_of_parameters))
        self._connection.txn_update()
        return result

    def executemany2(self, operation, seq_of_parameters):
        result = _invoke_with_reconnect(self.get_connection(), self,
                                               lambda: self._get_delegator().executemany2(operation, seq_of_parameters))
        self._connection.txn_update()
        return result

    def fetchone(self):
        return self._get_delegator().fetchone()

    def fetchone2(self):
        return self._get_delegator().fetchone2()

    def fetchmany(self, size=None):
        return self._get_delegator().fetchmany(size)

    def fetchall(self):
        return self._get_delegator().fetchall()

    def setinputsizes(self, sizes):
        self._get_delegator().setinputsizes(sizes)

    def setoutputsize(self, size, column=None):
        self._get_delegator().setoutputsize(size, column)

    @property
    def connection(self):
        return self._connection

    @property
    def rowcount(self):
        return self._get_delegator().rowcount

    @property
    def rownumber(self):
        return self._get_delegator().rownumber


class ServiceConnection(MultiConnection, ServiceConnectionI):

    def __init__(self, url, info):
        super().__init__(url, info)

    def unset_do_reconnect(self):
        with self._do_reconnect_lock:
            self._qe_address = None
            self._do_reconnect = False
            self._force_reopen = False

    def qe_addresses(self):
        res = _invoke_with_reconnect(self, None, lambda: self._delegator.serve_lx_action('QE_ADDRESSES'))
        if res:
            error = res.pop('ERROR', None)
            if error is None:
                return res['RESULT'].split(',')
            raise RuntimeError('Cannot get QE addresses: ' + error)
        raise IOError('Cannot get QE addresses')



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" zope.interface.verify unit tests
"""
import unittest

# pylint:disable=inherit-non-class,no-method-argument,no-self-argument

class Test_verifyClass(unittest.TestCase):

    verifier = None

    def setUp(self):
        self.verifier = self._get_FUT()

    @classmethod
    def _get_FUT(cls):
        from ....zope.interface.verify import verifyClass
        return verifyClass

    _adjust_object_before_verify = lambda self, x: x

    def _callFUT(self, iface, klass, **kwargs):
        return self.verifier(iface,
                             self._adjust_object_before_verify(klass),
                             **kwargs)

    def test_class_doesnt_implement(self):
        from ....zope.interface import Interface
        from ....zope.interface.exceptions import DoesNotImplement

        class ICurrent(Interface):
            pass

        class Current:
            pass

        self.assertRaises(DoesNotImplement, self._callFUT, ICurrent, Current)

    def test_class_doesnt_implement_but_classImplements_later(self):
        from ....zope.interface import Interface
        from ....zope.interface import classImplements

        class ICurrent(Interface):
            pass

        class Current:
            pass

        classImplements(Current, ICurrent)

        self._callFUT(ICurrent, Current)

    def test_class_doesnt_have_required_method_simple(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenImplementation

        class ICurrent(Interface):
            def method():
                """docstring"""

        @implementer(ICurrent)
        class Current:
            pass

        self.assertRaises(BrokenImplementation,
                          self._callFUT, ICurrent, Current)

    def test_class_has_required_method_simple(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):
            def method():
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self):
                raise NotImplementedError()

        self._callFUT(ICurrent, Current)

    def test_class_doesnt_have_required_method_derived(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenImplementation

        class IBase(Interface):
            def method():
                """docstring"""

        class IDerived(IBase):
            pass

        @implementer(IDerived)
        class Current:
            pass

        self.assertRaises(BrokenImplementation,
                          self._callFUT, IDerived, Current)

    def test_class_has_required_method_derived(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class IBase(Interface):
            def method():
                """docstring"""

        class IDerived(IBase):
            pass

        @implementer(IDerived)
        class Current:

            def method(self):
                raise NotImplementedError()

        self._callFUT(IDerived, Current)

    def test_method_takes_wrong_arg_names_but_OK(self):
        # We no longer require names to match.
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):

            def method(a):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, b):
                raise NotImplementedError()

        self._callFUT(ICurrent, Current)

    def test_method_takes_not_enough_args(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenMethodImplementation

        class ICurrent(Interface):

            def method(a):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self):
                raise NotImplementedError()

        self.assertRaises(BrokenMethodImplementation,
                          self._callFUT, ICurrent, Current)

    def test_method_doesnt_take_required_starargs(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenMethodImplementation

        class ICurrent(Interface):

            def method(*args):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self):
                raise NotImplementedError()

        self.assertRaises(BrokenMethodImplementation,
                          self._callFUT, ICurrent, Current)

    def test_method_doesnt_take_required_only_kwargs(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenMethodImplementation

        class ICurrent(Interface):

            def method(**kw):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self):
                raise NotImplementedError()

        self.assertRaises(BrokenMethodImplementation,
                          self._callFUT, ICurrent, Current)

    def test_method_takes_extra_arg(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenMethodImplementation

        class ICurrent(Interface):

            def method(a):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, a, b):
                raise NotImplementedError()

        self.assertRaises(BrokenMethodImplementation,
                          self._callFUT, ICurrent, Current)

    def test_method_takes_extra_arg_with_default(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):

            def method(a):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, a, b=None):
                raise NotImplementedError()

        self._callFUT(ICurrent, Current)

    def test_method_takes_only_positional_args(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):

            def method(a):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, *args):
                raise NotImplementedError()

        self._callFUT(ICurrent, Current)

    def test_method_takes_only_kwargs(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenMethodImplementation

        class ICurrent(Interface):

            def method(a):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, **kw):
                raise NotImplementedError()

        self.assertRaises(BrokenMethodImplementation,
                          self._callFUT, ICurrent, Current)

    def test_method_takes_extra_starargs(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):

            def method(a):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, a, *args):
                raise NotImplementedError()

        self._callFUT(ICurrent, Current)

    def test_method_takes_extra_starargs_and_kwargs(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):

            def method(a):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, a, *args, **kw):
                raise NotImplementedError()

        self._callFUT(ICurrent, Current)

    def test_method_doesnt_take_required_positional_and_starargs(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenMethodImplementation

        class ICurrent(Interface):

            def method(a, *args):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, a):
                raise NotImplementedError()

        self.assertRaises(BrokenMethodImplementation,
                          self._callFUT, ICurrent, Current)

    def test_method_takes_required_positional_and_starargs(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):

            def method(a, *args):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, a, *args):
                raise NotImplementedError()

        self._callFUT(ICurrent, Current)

    def test_method_takes_only_starargs(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):

            def method(a, *args):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, *args):
                raise NotImplementedError()

        self._callFUT(ICurrent, Current)

    def test_method_takes_required_kwargs(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):

            def method(**kwargs):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, **kw):
                raise NotImplementedError()

        self._callFUT(ICurrent, Current)

    def test_method_takes_positional_plus_required_starargs(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenMethodImplementation

        class ICurrent(Interface):

            def method(*args):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, a, *args):
                raise NotImplementedError()

        self.assertRaises(BrokenMethodImplementation,
                          self._callFUT, ICurrent, Current)


    def test_method_doesnt_take_required_kwargs(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenMethodImplementation

        class ICurrent(Interface):

            def method(**kwargs):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            def method(self, a):
                raise NotImplementedError()

        self.assertRaises(BrokenMethodImplementation,
                          self._callFUT, ICurrent, Current)


    def test_class_has_method_for_iface_attr(self):
        from ....zope.interface import Attribute
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):
            attr = Attribute("The foo Attribute")

        @implementer(ICurrent)
        class Current:

            def attr(self):
                raise NotImplementedError()

        self._callFUT(ICurrent, Current)

    def test_class_has_nonmethod_for_method(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenMethodImplementation

        class ICurrent(Interface):
            def method():
                """docstring"""

        @implementer(ICurrent)
        class Current:
            method = 1

        self.assertRaises(BrokenMethodImplementation,
                          self._callFUT, ICurrent, Current)

    def test_class_has_attribute_for_attribute(self):
        from ....zope.interface import Attribute
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):
            attr = Attribute("The foo Attribute")

        @implementer(ICurrent)
        class Current:

            attr = 1

        self._callFUT(ICurrent, Current)

    def test_class_misses_attribute_for_attribute(self):
        # This check *passes* for verifyClass
        from ....zope.interface import Attribute
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class ICurrent(Interface):
            attr = Attribute("The foo Attribute")

        @implementer(ICurrent)
        class Current:
            pass

        self._callFUT(ICurrent, Current)

    def test_w_callable_non_func_method(self):
        from ....zope.interface.interface import Method
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        class QuasiMethod(Method):
            def __call__(self, *args, **kw):
                raise NotImplementedError()

        class QuasiCallable:
            def __call__(self, *args, **kw):
                raise NotImplementedError()

        class ICurrent(Interface):
            attr = QuasiMethod('This is callable')

        @implementer(ICurrent)
        class Current:
            attr = QuasiCallable()

        self._callFUT(ICurrent, Current)


    def test_w_decorated_method(self):
        from ....zope.interface import Interface
        from ....zope.interface import implementer

        def decorator(func):
            # this is, in fact, zope.proxy.non_overridable
            return property(lambda self: func.__get__(self))

        class ICurrent(Interface):

            def method(a):
                """docstring"""

        @implementer(ICurrent)
        class Current:

            @decorator
            def method(self, a):
                raise NotImplementedError()

        self._callFUT(ICurrent, Current)

    def test_dict_IFullMapping(self):
        # A dict should be an IFullMapping, but this exposes two
        # issues. First, on CPython, methods of builtin types are
        # "method_descriptor" objects, and are harder to introspect.
        # Second, on PyPy, the signatures can be just plain wrong,
        # specifying as required arguments that are actually optional.
        # See https://github.com/zopefoundation/zope.interface/issues/118
        from ....zope.interface.common.mapping import IFullMapping
        self._callFUT(IFullMapping, dict, tentative=True)

    def test_list_ISequence(self):
        # As for test_dict_IFullMapping
        from ....zope.interface.common.sequence import ISequence
        self._callFUT(ISequence, list, tentative=True)

    def test_tuple_IReadSequence(self):
        # As for test_dict_IFullMapping
        from ....zope.interface.common.sequence import IReadSequence
        self._callFUT(IReadSequence, tuple, tentative=True)


    def test_multiple_invalid(self):
        from ....zope.interface.exceptions import MultipleInvalid
        from ....zope.interface.exceptions import DoesNotImplement
        from ....zope.interface.exceptions import BrokenImplementation
        from ....zope.interface import Interface
        from ....zope.interface import classImplements

        class ISeveralMethods(Interface):
            def meth1(arg1):
                "Method 1"
            def meth2(arg1):
                "Method 2"

        class SeveralMethods:
            pass

        with self.assertRaises(MultipleInvalid) as exc:
            self._callFUT(ISeveralMethods, SeveralMethods)

        ex = exc.exception
        self.assertEqual(3, len(ex.exceptions))
        self.assertIsInstance(ex.exceptions[0], DoesNotImplement)
        self.assertIsInstance(ex.exceptions[1], BrokenImplementation)
        self.assertIsInstance(ex.exceptions[2], BrokenImplementation)

        # If everything else is correct, only the single error is raised without
        # the wrapper.
        classImplements(SeveralMethods, ISeveralMethods)
        SeveralMethods.meth1 = lambda self, arg1: "Hi"

        with self.assertRaises(BrokenImplementation):
            self._callFUT(ISeveralMethods, SeveralMethods)

class Test_verifyObject(Test_verifyClass):

    @classmethod
    def _get_FUT(cls):
        from ....zope.interface.verify import verifyObject
        return verifyObject

    def _adjust_object_before_verify(self, target):
        if isinstance(target, (type, type(OldSkool))):
            target = target()
        return target

    def test_class_misses_attribute_for_attribute(self):
        # This check *fails* for verifyObject
        from ....zope.interface import Attribute
        from ....zope.interface import Interface
        from ....zope.interface import implementer
        from ....zope.interface.exceptions import BrokenImplementation

        class ICurrent(Interface):
            attr = Attribute("The foo Attribute")

        @implementer(ICurrent)
        class Current:
            pass

        self.assertRaises(BrokenImplementation,
                          self._callFUT, ICurrent, Current)

    def test_module_hit(self):
        from ....zope.interface.tests.idummy import IDummyModule
        from ....zope.interface.tests import dummy

        self._callFUT(IDummyModule, dummy)

    def test_module_miss(self):
        from ....zope.interface import Interface
        from ....zope.interface.tests import dummy
        from ....zope.interface.exceptions import DoesNotImplement

        # same name, different object
        class IDummyModule(Interface):
            pass

        self.assertRaises(DoesNotImplement,
                          self._callFUT, IDummyModule, dummy)

    def test_staticmethod_hit_on_class(self):
        from ....zope.interface import Interface
        from ....zope.interface import provider
        from ....zope.interface.verify import verifyObject

        class IFoo(Interface):

            def bar(a, b):
                "The bar method"

        @provider(IFoo)
        class Foo:

            @staticmethod
            def bar(a, b):
                raise AssertionError("We're never actually called")

        # Don't use self._callFUT, we don't want to instantiate the
        # class.
        verifyObject(IFoo, Foo)

class OldSkool:
    pass



# Copyright 2015 Lukas Lalinsky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import base64
import datetime
from decimal import Decimal
from ._vendor.DateTime.DateTime import long
from ._vendor import pytz
from lxdbapi.common_pb2 import Rep
from lxdbapi.errors import ArrayTypeException
from lxdbapi import common_pb2

__all__ = [
    'Date', 'Time', 'Timestamp', 'DateFromTicks', 'TimeFromTicks', 'TimestampFromTicks',
    'Binary', 'STRING', 'BINARY', 'NUMBER', 'DATETIME', 'ROWID', 'BOOLEAN', 'javaTypetoNative',
    'nativeToParamType', 'datetimeToTimezoneUtc', 'timezoneFromDatetimeUtc', 'LXARRAY',
    'typed_value_2_native', 'native_2_typed_value'
]


def Date(year, month, day):
    """Constructs an object holding a date value."""
    return datetime.date(year, month, day)


def Time(hour, minute, second):
    """Constructs an object holding a time value."""
    return datetime.time(hour, minute, second)


def Timestamp(year, month, day, hour, minute, second):
    """Constructs an object holding a datetime/timestamp value."""
    return datetime.datetime(year, month, day, hour, minute, second)


def DateFromTicks(ticks):
    """Constructs an object holding a date value from the given UNIX timestamp."""
    return Date(*time.localtime(ticks)[:3])


def TimeFromTicks(ticks):
    """Constructs an object holding a time value from the given UNIX timestamp."""
    return Time(*time.localtime(ticks)[3:6])


def TimestampFromTicks(ticks):
    """Constructs an object holding a datetime/timestamp value from the given UNIX timestamp."""
    return Timestamp(*time.localtime(ticks)[:6])


def Binary(value):
    """Constructs an object capable of holding a binary (long) string value."""
    if isinstance(value, _BinaryString):
        return value
    return _BinaryString(base64.b64encode(value))

def time_from_java_sql_time(n):
    dt = datetime.datetime(1970, 1, 1) + datetime.timedelta(milliseconds=n)
    return dt.time()


def time_to_java_sql_time(t):
    return long(((t.hour * 60 + t.minute) * 60 + t.second) * 1000 + t.microsecond / 1000)


def date_from_java_sql_date(n):
    return datetime.date(1970, 1, 1) + datetime.timedelta(days=n)


def date_to_java_sql_date(d):
    if isinstance(d, datetime.datetime):
        d = d.date()
    td = d - datetime.date(1970, 1, 1)
    return td.days


def datetime_from_java_sql_timestamp(n):
    return datetime.datetime(1970, 1, 1).replace(tzinfo=pytz.utc) + datetime.timedelta(milliseconds=n)


def datetime_to_java_sql_timestamp(d):
    td = d - datetime.datetime(1970, 1, 1).replace(tzinfo=pytz.utc)
    return long(td.microseconds / 1000 + (td.seconds + td.days * 24 * 3600) * 1000)

class _BinaryString(str):
    pass


class ColumnType(object):

    def __init__(self, eq_types):
        self.eq_types = tuple(eq_types)
        self.eq_types_set = set(eq_types)

    def __cmp__(self, other):
        if other in self.eq_types_set:
            return 0
        if other < self.eq_types:
            return 1
        else:
            return -1


STRING = ColumnType(['VARCHAR', 'CHAR'])
"""Type object that can be used to describe string-based columns."""

BINARY = ColumnType(['BINARY', 'VARBINARY'])
"""Type object that can be used to describe (long) binary columns."""

NUMBER = ColumnType(['INTEGER', 'UNSIGNED_INT', 'BIGINT', 'UNSIGNED_LONG', 'TINYINT', 'UNSIGNED_TINYINT', 'SMALLINT', 'UNSIGNED_SMALLINT', 'FLOAT', 'UNSIGNED_FLOAT', 'DOUBLE', 'UNSIGNED_DOUBLE', 'DECIMAL'])
"""Type object that can be used to describe numeric columns."""

DATETIME = ColumnType(['TIME', 'DATE', 'TIMESTAMP', 'UNSIGNED_TIME', 'UNSIGNED_DATE', 'UNSIGNED_TIMESTAMP'])
"""Type object that can be used to describe date/time columns."""

ROWID = ColumnType([])
"""Only implemented for DB API 2.0 compatibility, not used."""

BOOLEAN = ColumnType(['BOOLEAN'])
"""Type object that can be used to describe boolean columns. This is a lxdbapi-specific extension."""

# XXX ARRAY
class LXARRAY():
    __value = []
    __type = None
    def __init__(self, value, type):
        self.__value = value
        self.__type = type

    def getvalue(self):
        return self.__value
    def setvalue(self, value):
        self.__value = value

    def gettype(self):
        return self.__type

    def settype(self, type):
        self.__type = type


def typedValueToNative(v):
    if Rep.Name(v.type) == "BOOLEAN" or Rep.Name(v.type) == "PRIMITIVE_BOOLEAN":
        return v.bool_value

    elif Rep.Name(v.type) == "STRING" or Rep.Name(v.type) == "PRIMITIVE_CHAR" or Rep.Name(v.type) == "CHARACTER" or Rep.Name(v.type) == "BIG_DECIMAL":
        return v.string_value

    elif Rep.Name(v.type) == "FLOAT" or Rep.Name(v.type) == "PRIMITIVE_FLOAT" or Rep.Name(v.type) == "DOUBLE" or Rep.Name(v.type) == "PRIMITIVE_DOUBLE":
        return v.double_value

    elif Rep.Name(v.type) == "LONG" or Rep.Name(v.type) == "PRIMITIVE_LONG" or Rep.Name(v.type) == "INTEGER" or Rep.Name(v.type) == "PRIMITIVE_INT" or \
                    Rep.Name(v.type) == "BIG_INTEGER" or Rep.Name(v.type) == "NUMBER" or Rep.Name(v.type) == "BYTE" or Rep.Name(v.type) == "PRIMITIVE_BYTE" or \
                    Rep.Name(v.type) == "SHORT" or Rep.Name(v.type) == "PRIMITIVE_SHORT":
        return v.number_value

    elif Rep.Name(v.type) == "BYTE_STRING":
        return v.bytes_value

    else:
        return None

def javaTypetoNative(java_type):
    if java_type == 'java.math.BigDecimal':
        return ('BIG_DECIMAL', None, "string_value")
    elif java_type == 'java.lang.Float':
        return ('FLOAT', float, "double_value")
    elif java_type == 'java.lang.Double':
        return ('DOUBLE', None, "double_value")
    elif java_type == 'java.lang.Long':
        return ('LONG', None, "number_value")
    elif java_type == 'java.lang.Integer':
        return ('INTEGER', int, "number_value")
    elif java_type == 'java.lang.Short':
        return ('SHORT', int, "number_value")
    elif java_type == 'java.lang.Byte':
        return ('BYTE', Binary, "bytes_value")
    elif java_type == 'java.lang.Boolean':
        return ('BOOLEAN', bool, "bool_value")
    elif java_type == 'java.lang.String':
        return ('STRING', None, "string_value")
    elif java_type == 'java.sql.Time':
        return ('JAVA_SQL_TIME', time_from_java_sql_time, "number_value")
    elif java_type == 'java.sql.Date':
        return ('JAVA_SQL_DATE', date_from_java_sql_date, "number_value")
    elif java_type == 'java.sql.Timestamp':
        return ('JAVA_SQL_TIMESTAMP', datetime_from_java_sql_timestamp, "number_value")
    elif java_type == '[B':
        return ('BYTE_STRING', bytes, "bytes_value")
    else:
        return ('NULL', None)

def nativeToParamType(value, isArray=False):
    if isinstance(value, int):
        if(abs(value) <= 0x7FFFFFFF and not isArray):
            return ('INTEGER', int, "number_value")
        else:
            return ('LONG', int, "number_value")
    elif isinstance(value, float):
        return ('DOUBLE', float, "double_value")
    elif isinstance(value, bool):
        return ('BOOLEAN', bool, "bool_value")
    elif isinstance(value, str):
        return ('STRING', str, "string_value")
    elif isinstance(value, datetime.time):
        return ('JAVA_SQL_TIME', time_to_java_sql_time, "number_value")
    elif isinstance(value, datetime.datetime):  # TODO check: important to put this case before datetime.date as datetimes also match in date clause
        return ('JAVA_SQL_TIMESTAMP', datetime_to_java_sql_timestamp, "number_value")
    elif isinstance(value, datetime.date):
        return ('JAVA_SQL_DATE', date_to_java_sql_date, "number_value")
    elif isinstance(value, bytes):
        return ('BYTE_STRING', bytes, "bytes_value")
    elif isinstance(value, LXARRAY):
        return ('ARRAY', None, "array_value")
    elif value == None:
        return ('NULL', None, "null")
    else:
        return ('STRING', str, "string_value")

def datetimeToTimezoneUtc(param, useTZ):
    if isinstance(param, datetime.datetime):
        if useTZ:  # assume localtime from client and convert to UTC
            local_tz = datetime.datetime.now().astimezone().tzinfo
            param = param.replace(tzinfo=local_tz).astimezone(pytz.utc)
        else:  # assume it is UTC, dont do anything
            param = pytz.utc.localize(param, is_dst=None)
        return param

def timezoneFromDatetimeUtc(param, useTZ):
    if useTZ:
        local_tz = datetime.datetime.now().astimezone().tzinfo
        param = param.astimezone(local_tz)
    return param.replace(tzinfo=None)

def typed_value_2_native(typed_value: common_pb2.TypedValue):
    func = _rep_to_native_funcs[_rep_index_to_name[typed_value.type]]
    if func is None:
        _unsupported_typed_value(typed_value)
        return None
    return func(typed_value)

def _bool_value_2_native(typed_value: common_pb2.TypedValue):
    return bool(typed_value.bool_value)


def _string_value_2_native(typed_value: common_pb2.TypedValue):
    return str(typed_value.string_value)

def _big_decimal_value_2_native(typed_value: common_pb2.TypedValue):
    value = str(typed_value.string_value)
    if '.' not in value and ',' not in value:
        return int(value)
    return Decimal(value)

def _bytes_value_2_native(typed_value: common_pb2.TypedValue):
    return bytes(typed_value.bytes_value)


def _number_value_2_native(typed_value: common_pb2.TypedValue):
    return int(typed_value.number_value)


def _double_value_2_native(typed_value: common_pb2.TypedValue):
    return float(typed_value.double_value)


def _null_value_2_native(typed_value: common_pb2.TypedValue):
    return None


def _unsupported_typed_value(typed_value: common_pb2.TypedValue):
    raise NotImplementedError("Not supported: " + _rep_index_to_name[typed_value.type])


_rep_index_to_name = ['PRIMITIVE_BOOLEAN',
                      'PRIMITIVE_BYTE',
                      'PRIMITIVE_CHAR',
                      'PRIMITIVE_SHORT',
                      'PRIMITIVE_INT',
                      'PRIMITIVE_LONG',
                      'PRIMITIVE_FLOAT',
                      'PRIMITIVE_DOUBLE',
                      'BOOLEAN',
                      'BYTE',
                      'CHARACTER',
                      'SHORT',
                      'INTEGER',
                      'LONG',
                      'FLOAT',
                      'DOUBLE',
                      'JAVA_SQL_TIME',
                      'JAVA_SQL_TIMESTAMP',
                      'JAVA_SQL_DATE',
                      'JAVA_UTIL_DATE',
                      'BYTE_STRING',
                      'STRING',
                      'NUMBER',
                      'OBJECT',
                      'NULL',
                      'BIG_INTEGER',
                      'BIG_DECIMAL',
                      'ARRAY',
                      'STRUCT',
                      'MULTISET']


_rep_to_native_funcs = {'ARRAY': _unsupported_typed_value,
                        'BIG_DECIMAL': _big_decimal_value_2_native,
                        'BIG_INTEGER': _bytes_value_2_native,
                        'BOOLEAN': _bool_value_2_native,
                        'BYTE': _number_value_2_native,
                        'BYTE_STRING': _bytes_value_2_native,
                        'CHARACTER': _string_value_2_native,
                        'DOUBLE': _double_value_2_native,
                        'FLOAT': _double_value_2_native,
                        'INTEGER': _number_value_2_native,
                        'JAVA_SQL_DATE': _number_value_2_native,
                        'JAVA_SQL_TIME': _number_value_2_native,
                        'JAVA_SQL_TIMESTAMP': _number_value_2_native,
                        'JAVA_UTIL_DATE': _number_value_2_native,
                        'LONG': _number_value_2_native,
                        'MULTISET': _unsupported_typed_value,
                        'NULL': _null_value_2_native,
                        'NUMBER': _number_value_2_native,
                        'OBJECT': _unsupported_typed_value,
                        'PRIMITIVE_BOOLEAN': _bool_value_2_native,
                        'PRIMITIVE_BYTE': _number_value_2_native,
                        'PRIMITIVE_CHAR': _string_value_2_native,
                        'PRIMITIVE_DOUBLE': _number_value_2_native,
                        'PRIMITIVE_FLOAT': _double_value_2_native,
                        'PRIMITIVE_INT': _number_value_2_native,
                        'PRIMITIVE_LONG': _number_value_2_native,
                        'PRIMITIVE_SHORT': _number_value_2_native,
                        'SHORT': _number_value_2_native,
                        'STRING': _string_value_2_native,
                        'STRUCT': _unsupported_typed_value}


def native_2_typed_value(value):
    type_param = nativeToParamType(value)
    if value is None:
        result = common_pb2.TypedValue(null=True, type=common_pb2.Rep.Value('NULL'))
    elif isinstance(value, LXARRAY):

        result = common_pb2.TypedValue()  # parent typed value
        result.type = common_pb2.Rep.Value(type_param[0])
        if value.gettype() is int:
            result.component_type = common_pb2.Rep.Value("LONG")
        elif value.gettype() is float:
            result.component_type = common_pb2.Rep.Value("DOUBLE")
        elif value.gettype() is str:
            result.component_type = common_pb2.Rep.Value("STRING")
        else:
            raise ArrayTypeException("Array cannot be of type: {}".format(value.gettype()))

        for item in value.getvalue():
            type_param = nativeToParamType(item, isArray=True)  # int bigint varchar ???
            v_temp = common_pb2.TypedValue()
            v_temp.type = common_pb2.Rep.Value(type_param[0])
            if item != None:
                setattr(v_temp, type_param[2], type_param[1](item))
            else:
                setattr(v_temp, type_param[2], 1)  # 1 for True
            result.array_value.append(v_temp)
        return result
    else:
        result = common_pb2.TypedValue()
        result.type = common_pb2.Rep.Value(type_param[0])
        if type_param[1] is not None:
            setattr(result, type_param[2], type_param[1](value))
        else:
            setattr(result, type_param[2], value)
    return result



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from datetime import date, datetime, tzinfo, timedelta
import math
import platform
import os
import sys
import time
import unittest

from ... import pytz

from ...DateTime.DateTime import _findLocalTimeZoneName
from ...DateTime import DateTime

if sys.version_info > (3, ):
    import pickle
    unicode = str
    PY3K = True
else:
    import cPickle as pickle
    PY3K = False

try:
    __file__
except NameError:
    f = sys.argv[0]
else:
    f = __file__

IS_PYPY = getattr(platform, 'python_implementation', lambda: None)() == 'PyPy'

DATADIR = os.path.dirname(os.path.abspath(f))
del f

ZERO = timedelta(0)


class FixedOffset(tzinfo):
    """Fixed offset in minutes east from UTC."""

    def __init__(self, offset, name):
        self.__offset = timedelta(minutes=offset)
        self.__name = name

    def utcoffset(self, dt):
        return self.__offset

    def tzname(self, dt):
        return self.__name

    def dst(self, dt):
        return ZERO


class DateTimeTests(unittest.TestCase):

    def _compare(self, dt1, dt2):
        '''Compares the internal representation of dt1 with
        the representation in dt2.  Allows sub-millisecond variations.
        Primarily for testing.'''
        self.assertEqual(round(dt1._t, 3), round(dt2._t, 3))
        self.assertEqual(round(dt1._d, 9), round(dt2._d, 9))
        self.assertEqual(round(dt1.time, 9), round(dt2.time, 9))
        self.assertEqual(dt1.millis(), dt2.millis())
        self.assertEqual(dt1._micros, dt2._micros)

    def testBug1203(self):
        # 01:59:60 occurred in old DateTime
        dt = DateTime(7200, 'GMT')
        self.assertTrue(str(dt).find('60') < 0, dt)

    def testDSTInEffect(self):
        # Checks GMT offset for a DST date in the US/Eastern time zone
        dt = DateTime(2000, 5, 9, 15, 0, 0, 'US/Eastern')
        self.assertEqual(dt.toZone('GMT').hour(), 19,
                         (dt, dt.toZone('GMT')))

    def testDSTNotInEffect(self):
        # Checks GMT offset for a non-DST date in the US/Eastern time zone
        dt = DateTime(2000, 11, 9, 15, 0, 0, 'US/Eastern')
        self.assertEqual(dt.toZone('GMT').hour(), 20,
                         (dt, dt.toZone('GMT')))

    def testAddPrecision(self):
        # Precision of serial additions
        dt = DateTime()
        self.assertEqual(str(dt + 0.10 + 3.14 + 6.76 - 10), str(dt),
                         dt)

    def testConstructor3(self):
        # Constructor from date/time string
        dt = DateTime()
        dt1s = '%d/%d/%d %d:%d:%f %s' % (
            dt.year(),
            dt.month(),
            dt.day(),
            dt.hour(),
            dt.minute(),
            dt.second(),
            dt.timezone())
        dt1 = DateTime(dt1s)
        # Compare representations as it's the
        # only way to compare the dates to the same accuracy
        self.assertEqual(repr(dt), repr(dt1))

    def testConstructor4(self):
        # Constructor from time float
        dt = DateTime()
        dt1 = DateTime(float(dt))
        self._compare(dt, dt1)

    def testConstructor5(self):
        # Constructor from time float and timezone
        dt = DateTime()
        dt1 = DateTime(float(dt), dt.timezone())
        self.assertEqual(str(dt), str(dt1), (dt, dt1))
        dt1 = DateTime(float(dt), unicode(dt.timezone()))
        self.assertEqual(str(dt), str(dt1), (dt, dt1))

    def testConstructor6(self):
        # Constructor from year and julian date
        # This test must normalize the time zone, or it *will* break when
        # DST changes!
        dt1 = DateTime(2000, 5.500000578705)
        dt = DateTime('2000/1/5 12:00:00.050 pm %s' % dt1.localZone())
        self._compare(dt, dt1)

    def testConstructor7(self):
        # Constructor from parts
        dt = DateTime()
        dt1 = DateTime(
            dt.year(),
            dt.month(),
            dt.day(),
            dt.hour(),
            dt.minute(),
            dt.second(),
            dt.timezone())
        # Compare representations as it's the
        # only way to compare the dates to the same accuracy
        self.assertEqual(repr(dt), repr(dt1))

    def testDayOfWeek(self):
        # Compare to the datetime.date value to make it locale independent
        expected = date(2000, 6, 16).strftime('%A')
        # strftime() used to always be passed a day of week of 0
        dt = DateTime('2000/6/16')
        s = dt.strftime('%A')
        self.assertEqual(s, expected, (dt, s))

    def testOldDate(self):
        # Fails when an 1800 date is displayed with negative signs
        dt = DateTime('1830/5/6 12:31:46.213 pm')
        dt1 = dt.toZone('GMT+6')
        self.assertTrue(str(dt1).find('-') < 0, (dt, dt1))

    def testSubtraction(self):
        # Reconstruction of a DateTime from its parts, with subtraction
        # this also tests the accuracy of addition and reconstruction
        dt = DateTime()
        dt1 = dt - 3.141592653
        dt2 = DateTime(
            dt.year(),
            dt.month(),
            dt.day(),
            dt.hour(),
            dt.minute(),
            dt.second())
        dt3 = dt2 - 3.141592653
        self.assertEqual(dt1, dt3, (dt, dt1, dt2, dt3))

    def testTZ1add(self):
        # Time zone manipulation: add to a date
        dt = DateTime('1997/3/8 1:45am GMT-4')
        dt1 = DateTime('1997/3/9 1:45pm GMT+8')
        self.assertTrue((dt + 1.0).equalTo(dt1))

    def testTZ1sub(self):
        # Time zone manipulation: subtract from a date
        dt = DateTime('1997/3/8 1:45am GMT-4')
        dt1 = DateTime('1997/3/9 1:45pm GMT+8')
        self.assertTrue((dt1 - 1.0).equalTo(dt))

    def testTZ1diff(self):
        # Time zone manipulation: diff two dates
        dt = DateTime('1997/3/8 1:45am GMT-4')
        dt1 = DateTime('1997/3/9 1:45pm GMT+8')
        self.assertEqual(dt1 - dt, 1.0, (dt, dt1))

    def test_compare_methods(self):
        # Compare two dates using several methods
        dt = DateTime('1997/1/1')
        dt1 = DateTime('1997/2/2')
        self.assertTrue(dt1.greaterThan(dt))
        self.assertTrue(dt1.greaterThanEqualTo(dt))
        self.assertTrue(dt.lessThan(dt1))
        self.assertTrue(dt.lessThanEqualTo(dt1))
        self.assertTrue(dt.notEqualTo(dt1))
        self.assertFalse(dt.equalTo(dt1))

    def test_compare_methods_none(self):
        # Compare a date to None
        dt = DateTime('1997/1/1')
        self.assertTrue(dt.greaterThan(None))
        self.assertTrue(dt.greaterThanEqualTo(None))
        self.assertFalse(dt.lessThan(None))
        self.assertFalse(dt.lessThanEqualTo(None))
        self.assertTrue(dt.notEqualTo(None))
        self.assertFalse(dt.equalTo(None))

    def test_pickle(self):
        dt = DateTime()
        data = pickle.dumps(dt, 1)
        new = pickle.loads(data)
        for key in DateTime.__slots__:
            self.assertEqual(getattr(dt, key), getattr(new, key))

    def test_pickle_with_tz(self):
        dt = DateTime('2002/5/2 8:00am GMT+8')
        data = pickle.dumps(dt, 1)
        new = pickle.loads(data)
        for key in DateTime.__slots__:
            self.assertEqual(getattr(dt, key), getattr(new, key))

    def test_pickle_with_numerical_tz(self):
        for dt_str in ('2007/01/02 12:34:56.789 +0300',
                       '2007/01/02 12:34:56.789 +0430',
                       '2007/01/02 12:34:56.789 -1234'):
            dt = DateTime(dt_str)
            data = pickle.dumps(dt, 1)
            new = pickle.loads(data)
            for key in DateTime.__slots__:
                self.assertEqual(getattr(dt, key), getattr(new, key))

    def test_pickle_with_micros(self):
        dt = DateTime('2002/5/2 8:00:14.123 GMT+8')
        data = pickle.dumps(dt, 1)
        new = pickle.loads(data)
        for key in DateTime.__slots__:
            self.assertEqual(getattr(dt, key), getattr(new, key))

    def test_pickle_old(self):
        dt = DateTime('2002/5/2 8:00am GMT+0')
        data = ('(cDateTime.DateTime\nDateTime\nq\x01Noq\x02}q\x03(U\x05'
            '_amonq\x04U\x03Mayq\x05U\x05_adayq\x06U\x03Thuq\x07U\x05_pmonq'
            '\x08h\x05U\x05_hourq\tK\x08U\x05_fmonq\nh\x05U\x05_pdayq\x0bU'
            '\x04Thu.q\x0cU\x05_fdayq\rU\x08Thursdayq\x0eU\x03_pmq\x0fU\x02amq'
            '\x10U\x02_tq\x11GA\xcehy\x00\x00\x00\x00U\x07_minuteq\x12K\x00U'
            '\x07_microsq\x13L1020326400000000L\nU\x02_dq\x14G@\xe2\x12j\xaa'
            '\xaa\xaa\xabU\x07_secondq\x15G\x00\x00\x00\x00\x00\x00\x00\x00U'
            '\x03_tzq\x16U\x05GMT+0q\x17U\x06_monthq\x18K\x05U'
            '\x0f_timezone_naiveq\x19I00\nU\x04_dayq\x1aK\x02U\x05_yearq'
            '\x1bM\xd2\x07U\x08_nearsecq\x1cG\x00\x00\x00\x00\x00\x00\x00'
            '\x00U\x07_pmhourq\x1dK\x08U\n_dayoffsetq\x1eK\x04U\x04timeq'
            '\x1fG?\xd5UUUV\x00\x00ub.')
        if PY3K:
            data = data.encode('latin-1')
        new = pickle.loads(data)
        for key in DateTime.__slots__:
            self.assertEqual(getattr(dt, key), getattr(new, key))

    def test_pickle_old_without_micros(self):
        dt = DateTime('2002/5/2 8:00am GMT+0')
        data = ('(cDateTime.DateTime\nDateTime\nq\x01Noq\x02}q\x03(U\x05'
            '_amonq\x04U\x03Mayq\x05U\x05_adayq\x06U\x03Thuq\x07U\x05_pmonq'
            '\x08h\x05U\x05_hourq\tK\x08U\x05_fmonq\nh\x05U\x05_pdayq\x0bU'
            '\x04Thu.q\x0cU\x05_fdayq\rU\x08Thursdayq\x0eU\x03_pmq\x0fU'
            '\x02amq\x10U\x02_tq\x11GA\xcehy\x00\x00\x00\x00U\x07_minuteq'
            '\x12K\x00U\x02_dq\x13G@\xe2\x12j\xaa\xaa\xaa\xabU\x07_secondq'
            '\x14G\x00\x00\x00\x00\x00\x00\x00\x00U\x03_tzq\x15U\x05GMT+0q'
            '\x16U\x06_monthq\x17K\x05U\x0f_timezone_naiveq\x18I00\nU'
            '\x04_dayq\x19K\x02U\x05_yearq\x1aM\xd2\x07U\x08_nearsecq'
            '\x1bG\x00\x00\x00\x00\x00\x00\x00\x00U\x07_pmhourq\x1cK\x08U'
            '\n_dayoffsetq\x1dK\x04U\x04timeq\x1eG?\xd5UUUV\x00\x00ub.')
        if PY3K:
            data = data.encode('latin-1')
        new = pickle.loads(data)
        for key in DateTime.__slots__:
            self.assertEqual(getattr(dt, key), getattr(new, key))

    def testTZ2(self):
        # Time zone manipulation test 2
        dt = DateTime()
        dt1 = dt.toZone('GMT')
        s = dt.second()
        s1 = dt1.second()
        self.assertEqual(s, s1, (dt, dt1, s, s1))

    def testTZDiffDaylight(self):
        # Diff dates across daylight savings dates
        dt = DateTime('2000/6/8 1:45am US/Eastern')
        dt1 = DateTime('2000/12/8 12:45am US/Eastern')
        self.assertEqual(dt1 - dt, 183, (dt, dt1, dt1 - dt))

    def testY10KDate(self):
        # Comparison of a Y10K date and a Y2K date
        dt = DateTime('10213/09/21')
        dt1 = DateTime(2000, 1, 1)

        dsec = (dt.millis() - dt1.millis()) / 1000.0
        ddays = math.floor((dsec / 86400.0) + 0.5)

        self.assertEqual(ddays, 3000000, ddays)

    def test_tzoffset(self):
        # Test time-zone given as an offset

        # GMT
        dt = DateTime('Tue, 10 Sep 2001 09:41:03 GMT')
        self.assertEqual(dt.tzoffset(), 0)

        # Timezone by name, a timezone that hasn't got daylightsaving.
        dt = DateTime('Tue, 2 Mar 2001 09:41:03 GMT+3')
        self.assertEqual(dt.tzoffset(), 10800)

        # Timezone by name, has daylightsaving but is not in effect.
        dt = DateTime('Tue, 21 Jan 2001 09:41:03 PST')
        self.assertEqual(dt.tzoffset(), -28800)

        # Timezone by name, with daylightsaving in effect
        dt = DateTime('Tue, 24 Aug 2001 09:41:03 PST')
        self.assertEqual(dt.tzoffset(), -25200)

        # A negative numerical timezone
        dt = DateTime('Tue, 24 Jul 2001 09:41:03 -0400')
        self.assertEqual(dt.tzoffset(), -14400)

        # A positive numerical timzone
        dt = DateTime('Tue, 6 Dec 1966 01:41:03 +0200')
        self.assertEqual(dt.tzoffset(), 7200)

        # A negative numerical timezone with minutes.
        dt = DateTime('Tue, 24 Jul 2001 09:41:03 -0637')
        self.assertEqual(dt.tzoffset(), -23820)

        # A positive numerical timezone with minutes.
        dt = DateTime('Tue, 24 Jul 2001 09:41:03 +0425')
        self.assertEqual(dt.tzoffset(), 15900)

    def testISO8601(self):
        # ISO8601 reference dates
        ref0 = DateTime('2002/5/2 8:00am GMT')
        ref1 = DateTime('2002/5/2 8:00am US/Eastern')
        ref2 = DateTime('2006/11/6 10:30 GMT')
        ref3 = DateTime('2004/06/14 14:30:15 GMT-3')
        ref4 = DateTime('2006/01/01 GMT')

        # Basic tests
        # Though this is timezone naive and according to specification should
        # be interpreted in the local timezone, to preserve backwards
        # compatibility with previously expected behaviour.
        isoDt = DateTime('2002-05-02T08:00:00')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('2002-05-02T08:00:00Z')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('2002-05-02T08:00:00+00:00')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('2002-05-02T08:00:00-04:00')
        self.assertTrue(ref1.equalTo(isoDt))
        isoDt = DateTime('2002-05-02 08:00:00-04:00')
        self.assertTrue(ref1.equalTo(isoDt))

        # Bug 1386: the colon in the timezone offset is optional
        isoDt = DateTime('2002-05-02T08:00:00-0400')
        self.assertTrue(ref1.equalTo(isoDt))

        # Bug 2191: date reduced formats
        isoDt = DateTime('2006-01-01')
        self.assertTrue(ref4.equalTo(isoDt))
        isoDt = DateTime('200601-01')
        self.assertTrue(ref4.equalTo(isoDt))
        isoDt = DateTime('20060101')
        self.assertTrue(ref4.equalTo(isoDt))
        isoDt = DateTime('2006-01')
        self.assertTrue(ref4.equalTo(isoDt))
        isoDt = DateTime('200601')
        self.assertTrue(ref4.equalTo(isoDt))
        isoDt = DateTime('2006')
        self.assertTrue(ref4.equalTo(isoDt))

        # Bug 2191: date/time separators are also optional
        isoDt = DateTime('20020502T08:00:00')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('2002-05-02T080000')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('20020502T080000')
        self.assertTrue(ref0.equalTo(isoDt))

        # Bug 2191: timezones with only one digit for hour
        isoDt = DateTime('20020502T080000+0')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('20020502 080000-4')
        self.assertTrue(ref1.equalTo(isoDt))
        isoDt = DateTime('20020502T080000-400')
        self.assertTrue(ref1.equalTo(isoDt))
        isoDt = DateTime('20020502T080000-4:00')
        self.assertTrue(ref1.equalTo(isoDt))

        # Bug 2191: optional seconds/minutes
        isoDt = DateTime('2002-05-02T0800')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('2002-05-02T08')
        self.assertTrue(ref0.equalTo(isoDt))

        # Bug 2191: week format
        isoDt = DateTime('2002-W18-4T0800')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('2002-W184T0800')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('2002W18-4T0800')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('2002W184T08')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('2004-W25-1T14:30:15-03:00')
        self.assertTrue(ref3.equalTo(isoDt))
        isoDt = DateTime('2004-W25T14:30:15-03:00')
        self.assertTrue(ref3.equalTo(isoDt))

        # Bug 2191: day of year format
        isoDt = DateTime('2002-122T0800')
        self.assertTrue(ref0.equalTo(isoDt))
        isoDt = DateTime('2002122T0800')
        self.assertTrue(ref0.equalTo(isoDt))

        # Bug 2191: hours/minutes fractions
        isoDt = DateTime('2006-11-06T10.5')
        self.assertTrue(ref2.equalTo(isoDt))
        isoDt = DateTime('2006-11-06T10,5')
        self.assertTrue(ref2.equalTo(isoDt))
        isoDt = DateTime('20040614T1430.25-3')
        self.assertTrue(ref3.equalTo(isoDt))
        isoDt = DateTime('2004-06-14T1430,25-3')
        self.assertTrue(ref3.equalTo(isoDt))
        isoDt = DateTime('2004-06-14T14:30.25-3')
        self.assertTrue(ref3.equalTo(isoDt))
        isoDt = DateTime('20040614T14:30,25-3')
        self.assertTrue(ref3.equalTo(isoDt))

        # ISO8601 standard format
        iso8601_string = '2002-05-02T08:00:00-04:00'
        iso8601DT = DateTime(iso8601_string)
        self.assertEqual(iso8601_string, iso8601DT.ISO8601())

        # ISO format with no timezone
        isoDt = DateTime('2006-01-01 00:00:00')
        self.assertTrue(ref4.equalTo(isoDt))

    def testJulianWeek(self):
        # Check JulianDayWeek function
        fn = os.path.join(DATADIR, 'julian_testdata.txt')
        with open(fn, 'r') as fd:
            lines = fd.readlines()
        for line in lines:
            d = DateTime(line[:10])
            result_from_mx = tuple(map(int, line[12:-2].split(',')))
            self.assertEqual(result_from_mx[1], d.week())

    def testCopyConstructor(self):
        d = DateTime('2004/04/04')
        self.assertEqual(DateTime(d), d)
        self.assertEqual(str(DateTime(d)), str(d))
        d2 = DateTime('1999/04/12 01:00:00')
        self.assertEqual(DateTime(d2), d2)
        self.assertEqual(str(DateTime(d2)), str(d2))

    def testCopyConstructorPreservesTimezone(self):
        # test for https://bugs.launchpad.net/zope2/+bug/200007
        # This always worked in the local timezone, so we need at least
        # two tests with different zones to be sure at least one of them
        # is not local.
        d = DateTime('2004/04/04')
        self.assertEqual(DateTime(d).timezone(), d.timezone())
        d2 = DateTime('2008/04/25 12:00:00 EST')
        self.assertEqual(DateTime(d2).timezone(), d2.timezone())
        self.assertEqual(str(DateTime(d2)), str(d2))
        d3 = DateTime('2008/04/25 12:00:00 PST')
        self.assertEqual(DateTime(d3).timezone(), d3.timezone())
        self.assertEqual(str(DateTime(d3)), str(d3))

    def testRFC822(self):
        # rfc822 conversion
        dt = DateTime('2002-05-02T08:00:00+00:00')
        self.assertEqual(dt.rfc822(), 'Thu, 02 May 2002 08:00:00 +0000')

        dt = DateTime('2002-05-02T08:00:00+02:00')
        self.assertEqual(dt.rfc822(), 'Thu, 02 May 2002 08:00:00 +0200')

        dt = DateTime('2002-05-02T08:00:00-02:00')
        self.assertEqual(dt.rfc822(), 'Thu, 02 May 2002 08:00:00 -0200')

        # Checking that conversion from local time is working.
        dt = DateTime()
        dts = dt.rfc822().split(' ')
        times = dts[4].split(':')
        _isDST = time.localtime(time.time())[8]
        if _isDST:
            offset = time.altzone
        else:
            offset = time.timezone
        self.assertEqual(dts[0], dt.aDay() + ',')
        self.assertEqual(int(dts[1]), dt.day())
        self.assertEqual(dts[2], dt.aMonth())
        self.assertEqual(int(dts[3]), dt.year())
        self.assertEqual(int(times[0]), dt.h_24())
        self.assertEqual(int(times[1]), dt.minute())
        self.assertEqual(int(times[2]), int(dt.second()))
        self.assertEqual(dts[5], "%+03d%02d" % divmod((-offset / 60), 60))

    def testInternationalDateformat(self):
        for year in (1990, 2001, 2020):
            for month in (1, 12):
                for day in (1, 12, 28, 31):
                    try:
                        d_us = DateTime("%d/%d/%d" % (year, month, day))
                    except Exception:
                        continue

                    d_int = DateTime("%d.%d.%d" % (day, month, year),
                                     datefmt="international")
                    self.assertEqual(d_us, d_int)

                    d_int = DateTime("%d/%d/%d" % (day, month, year),
                                     datefmt="international")
                    self.assertEqual(d_us, d_int)

    def test_intl_format_hyphen(self):
        d_jan = DateTime('2011-01-11 GMT')
        d_nov = DateTime('2011-11-01 GMT')
        d_us = DateTime('11-01-2011 GMT')
        d_int = DateTime('11-01-2011 GMT', datefmt="international")
        self.assertNotEqual(d_us, d_int)
        self.assertEqual(d_us, d_nov)
        self.assertEqual(d_int, d_jan)

    def test_calcTimezoneName(self):
        from ...DateTime.interfaces import TimeError
        timezone_dependent_epoch = 2177452800
        try:
            DateTime()._calcTimezoneName(timezone_dependent_epoch, 0)
        except TimeError:
            self.fail('Zope Collector issue #484 (negative time bug): '
                      'TimeError raised')

    def testStrftimeTZhandling(self):
        # strftime timezone testing
        # This is a test for collector issue #1127
        format = '%Y-%m-%d %H:%M %Z'
        dt = DateTime('Wed, 19 Nov 2003 18:32:07 -0215')
        dt_string = dt.strftime(format)
        dt_local = dt.toZone(_findLocalTimeZoneName(0))
        dt_localstring = dt_local.strftime(format)
        self.assertEqual(dt_string, dt_localstring)

    def testStrftimeFarDates(self):
        # Checks strftime in dates <= 1900 or >= 2038
        dt = DateTime('1900/01/30')
        self.assertEqual(dt.strftime('%d/%m/%Y'), '30/01/1900')
        dt = DateTime('2040/01/30')
        self.assertEqual(dt.strftime('%d/%m/%Y'), '30/01/2040')

    def testZoneInFarDates(self):
        # Checks time zone in dates <= 1900 or >= 2038
        dt1 = DateTime('2040/01/30 14:33 GMT+1')
        dt2 = DateTime('2040/01/30 11:33 GMT-2')
        self.assertEqual(dt1.strftime('%d/%m/%Y %H:%M'),
                         dt2.strftime('%d/%m/%Y %H:%M'))

    def testStrftimeUnicode(self):
        if IS_PYPY:
            # Using Non-Ascii characters for strftime doesn't work in PyPy
            # https://bitbucket.org/pypy/pypy/issues/2161/pypy3-strftime-does-not-accept-unicode
            return
        dt = DateTime('2002-05-02T08:00:00+00:00')
        uchar = b'\xc3\xa0'.decode('utf-8')
        ok = dt.strftime('Le %d/%m/%Y a %Hh%M').replace('a', uchar)
        ustr = b'Le %d/%m/%Y \xc3\xa0 %Hh%M'.decode('utf-8')
        self.assertEqual(dt.strftime(ustr), ok)

    def testTimezoneNaiveHandling(self):
        # checks that we assign timezone naivity correctly
        dt = DateTime('2007-10-04T08:00:00+00:00')
        self.assertFalse(dt.timezoneNaive(),
            'error with naivity handling in __parse_iso8601')
        dt = DateTime('2007-10-04T08:00:00Z')
        self.assertFalse(dt.timezoneNaive(),
            'error with naivity handling in __parse_iso8601')
        dt = DateTime('2007-10-04T08:00:00')
        self.assertTrue(dt.timezoneNaive(),
            'error with naivity handling in __parse_iso8601')
        dt = DateTime('2007/10/04 15:12:33.487618 GMT+1')
        self.assertFalse(dt.timezoneNaive(),
            'error with naivity handling in _parse')
        dt = DateTime('2007/10/04 15:12:33.487618')
        self.assertTrue(dt.timezoneNaive(),
            'error with naivity handling in _parse')
        dt = DateTime()
        self.assertFalse(dt.timezoneNaive(),
            'error with naivity for current time')
        s = '2007-10-04T08:00:00'
        dt = DateTime(s)
        self.assertEqual(s, dt.ISO8601())
        s = '2007-10-04T08:00:00+00:00'
        dt = DateTime(s)
        self.assertEqual(s, dt.ISO8601())

    def testConversions(self):
        sdt0 = datetime.now()  # this is a timezone naive datetime
        dt0 = DateTime(sdt0)
        self.assertTrue(dt0.timezoneNaive(), (sdt0, dt0))
        sdt1 = datetime(2007, 10, 4, 18, 14, 42, 580, pytz.utc)
        dt1 = DateTime(sdt1)
        self.assertFalse(dt1.timezoneNaive(), (sdt1, dt1))

        # convert back
        sdt2 = dt0.asdatetime()
        self.assertEqual(sdt0, sdt2)
        sdt3 = dt1.utcdatetime()  # this returns a timezone naive datetime
        self.assertEqual(sdt1.hour, sdt3.hour)

        dt4 = DateTime('2007-10-04T10:00:00+05:00')
        sdt4 = datetime(2007, 10, 4, 5, 0)
        self.assertEqual(dt4.utcdatetime(), sdt4)
        self.assertEqual(dt4.asdatetime(), sdt4.replace(tzinfo=pytz.utc))

        dt5 = DateTime('2007-10-23 10:00:00 US/Eastern')
        tz = pytz.timezone('US/Eastern')
        sdt5 = datetime(2007, 10, 23, 10, 0, tzinfo=tz)
        dt6 = DateTime(sdt5)
        self.assertEqual(dt5.asdatetime(), sdt5)
        self.assertEqual(dt6.asdatetime(), sdt5)
        self.assertEqual(dt5, dt6)
        self.assertEqual(dt5.asdatetime().tzinfo, tz)
        self.assertEqual(dt6.asdatetime().tzinfo, tz)

    def testBasicTZ(self):
        # psycopg2 supplies it's own tzinfo instances, with no `zone` attribute
        tz = FixedOffset(60, 'GMT+1')
        dt1 = datetime(2008, 8, 5, 12, 0, tzinfo=tz)
        DT = DateTime(dt1)
        dt2 = DT.asdatetime()
        offset1 = dt1.tzinfo.utcoffset(dt1)
        offset2 = dt2.tzinfo.utcoffset(dt2)
        self.assertEqual(offset1, offset2)

    def testEDTTimezone(self):
        # should be able to parse EDT timezones:  see lp:599856.
        dt = DateTime("Mon, 28 Jun 2010 10:12:25 EDT")
        self.assertEqual(dt.Day(), 'Monday')
        self.assertEqual(dt.day(), 28)
        self.assertEqual(dt.Month(), 'June')
        self.assertEqual(dt.timezone(), 'GMT-4')

    def testParseISO8601(self):
        parsed = DateTime()._parse_iso8601('2010-10-10')
        self.assertEqual(parsed, (2010, 10, 10, 0, 0, 0, 'GMT+0000'))

    def test_interface(self):
        from ...DateTime.interfaces import IDateTime
        self.assertTrue(IDateTime.providedBy(DateTime()))

    def test_security(self):
        dt = DateTime()
        self.assertEqual(dt.__roles__, None)
        self.assertEqual(dt.__allow_access_to_unprotected_subobjects__, 1)


def test_suite():
    import doctest
    return unittest.TestSuite([
        unittest.makeSuite(DateTimeTests),
        doctest.DocFileSuite('DateTime.txt', package='DateTime'),
        doctest.DocFileSuite('pytz.txt', package='DateTime'),
    ])



# Copyright 2015 Lukas Lalinsky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid
import weakref
from lxdbapi import errorsOld
from lxdbapi.client import OPEN_CONNECTION_PROPERTIES, ONLY_CLIENT_PROPERTIES
from lxdbapi.cursor import Cursor
from lxdbapi.errors import ProgrammingError

__all__ = ['Connection']

logger = logging.getLogger(__name__)


class Connection(object):
    """Database connection.

    You should not construct this object manually, use :func:`~connect` instead.
    """

    def __init__(self, client, conn_id=None, cursor_factory=None, **kwargs):
        if conn_id:
            self._id = conn_id
        else:
            self._id = str(uuid.uuid4())
        self._client = client
        self._closed = True
        if cursor_factory is not None:
            self.cursor_factory = cursor_factory
        else:
            self.cursor_factory = Cursor
        self._cursors = []
        # Extract properties to pass to OpenConnectionRequest
        self._connection_args = {}
        # The rest of the kwargs
        self._filtered_args = {}
        for k in kwargs:
            if k != 'secure':
                if k in OPEN_CONNECTION_PROPERTIES:
                    self._connection_args[k] = kwargs[k]
                elif k not in ONLY_CLIENT_PROPERTIES:
                    self._filtered_args[k] = kwargs[k]
        self._connection_args['is_kryo'] = 'false'
        self._client.set_avatica_con(self)

    def __del__(self):
        if not self._closed:
            try:
                self.close()
            except BaseException as ex:
                logger.debug("Ignoring exception while closing: {}".format(ex))
            finally:
                self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not self._closed:
            self.close()

    def open(self):  # call just once
        """Opens the connection."""
        self._client.connect()
        self._client.open_connection(self._id, info=self._connection_args)
        self._closed = False
        self.set_session(**self._filtered_args)

    def close(self):  # updated
        """Closes the connection.
        No further operations are allowed, either on the connection or any
        of its cursors, once the connection is closed.

        If the connection is used in a ``with`` statement, this method will
        be automatically called at the end of the ``with`` block.
        """
        try:
            if self._closed:
                raise ProgrammingError('the connection is already closed')
            for cursor_ref in self._cursors:
                cursor = cursor_ref()
                if cursor is not None and not cursor._closed:
                    cursor.close()
            self._client.close_connection(self._id)
        finally:
            self._client.close()
            self._closed = True

    @property
    def closed(self):
        """Read-only attribute specifying if the connection is closed or not."""
        return self._closed

    def commit(self):
        """Commits pending database changes.

        You need to use :attr:`autocommit` mode.
        """

        self._client.commit(self._id)  # call added commit in avatica
        # TODO can support be added for this?
        if self._closed:
            raise ProgrammingError('the connection is already closed')

    def rollback(self):
        """Rollbacks pending database changes.
        """

        self._client.rollback(self._id)  # call added rollback in avatica
        # TODO can support be added for this?
        if self._closed:
            raise ProgrammingError('the connection is already closed')

    def cursor(self, cursor_factory=None):
        """Creates a new cursor.

        :param cursor_factory:
            This argument can be used to create non-standard cursors.
            The class returned must be a subclass of
            :class:`~phoenixdb.cursor.Cursor` (for example :class:`~phoenixdb.cursor.DictCursor`).
            A default factory for the connection can also be specified using the
            :attr:`cursor_factory` attribute.

        :returns:
            A :class:`~phoenixdb.cursor.Cursor` object.
        """
        if self._closed:
            raise ProgrammingError('the connection is already closed')
        cursor = (cursor_factory or self.cursor_factory)(self)
        self._cursors.append(weakref.ref(cursor, self._cursors.remove))
        return cursor

    def set_session(self, autocommit=None, readonly=None, useTZ=False):
        """Sets one or more parameters in the current connection.

        :param autocommit:
            Switch the connection to autocommit mode. With the current
            version, you need to always enable this, because
            :meth:`commit` is not implemented.

        :param readonly:
            Switch the connection to read-only mode.

        :param useTZ:
            False by default. If useTZ, use localTime of client. Else, use UTC/GMT
        """
        props = {}
        if autocommit is not None:
            props['autoCommit'] = bool(autocommit)
        if readonly is not None:
            props['readOnly'] = bool(readonly)

        props = self._client.connection_sync(self._id, props)
        self._autocommit = props.auto_commit
        self._readonly = props.read_only
        self._transactionisolation = props.transaction_isolation
        self._useTZ = useTZ

    @property
    def autocommit(self):
        """Read/write attribute for switching the connection's autocommit mode."""
        return self._autocommit

    @autocommit.setter
    def autocommit(self, value):
        if self._closed:
            raise ProgrammingError('the connection is already closed')
        props = self._client.connectionSync(self._id, {'autoCommit': bool(value)})
        self._autocommit = props.auto_commit

    @property
    def readonly(self):
        """Read/write attribute for switching the connection's readonly mode."""
        return self._readonly

    @readonly.setter
    def readonly(self, value):
        if self._closed:
            raise ProgrammingError('the connection is already closed')
        props = self._client.connectionSync(self._id, {'readOnly': bool(value)})
        self._readonly = props.read_only

    @property
    def transactionisolation(self):
        return self._transactionisolation

    @property
    def useTZ(self):
        return self._useTZ

    @transactionisolation.setter
    def transactionisolation(self, value):
        if self._closed:
            raise ProgrammingError('the connection is already closed')
        props = self._client.connectionSync(self._id, {'transactionIsolation': bool(value)})
        self._transactionisolation = props.transaction_isolation

    @property
    def url(self):
        return self._client.url

    def serve_lx_action(self, action, params=None):
        return self._client.serve_lx_action(action, params)


for name in errorsOld.__all__:
    setattr(Connection, name, getattr(errorsOld, name))



##############################################################################
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
##############################################################################
"""Implementation of interface declarations

There are three flavors of declarations:

  - Declarations are used to simply name declared interfaces.

  - ImplementsDeclarations are used to express the interfaces that a
    class implements (that instances of the class provides).

    Implements specifications support inheriting interfaces.

  - ProvidesDeclarations are used to express interfaces directly
    provided by objects.

"""
__docformat__ = 'restructuredtext'

import sys
from types import FunctionType
from types import MethodType
from types import ModuleType
import weakref

from ...zope.interface.interface import Interface
from ...zope.interface.interface import InterfaceClass
from ...zope.interface.interface import SpecificationBase
from ...zope.interface.interface import Specification
from ...zope.interface.interface import NameAndModuleComparisonMixin
from ...zope.interface._compat import _use_c_impl

__all__ = [
    # None. The public APIs of this module are
    # re-exported from zope.interface directly.
]

# pylint:disable=too-many-lines

# Registry of class-implementation specifications
BuiltinImplementationSpecifications = {}


def _next_super_class(ob):
    # When ``ob`` is an instance of ``super``, return
    # the next class in the MRO that we should actually be
    # looking at. Watch out for diamond inheritance!
    self_class = ob.__self_class__
    class_that_invoked_super = ob.__thisclass__
    complete_mro = self_class.__mro__
    next_class = complete_mro[complete_mro.index(class_that_invoked_super) + 1]
    return next_class

class named:

    def __init__(self, name):
        self.name = name

    def __call__(self, ob):
        ob.__component_name__ = self.name
        return ob


class Declaration(Specification):
    """Interface declarations"""

    __slots__ = ()

    def __init__(self, *bases):
        Specification.__init__(self, _normalizeargs(bases))

    def __contains__(self, interface):
        """Test whether an interface is in the specification
        """

        return self.extends(interface) and interface in self.interfaces()

    def __iter__(self):
        """Return an iterator for the interfaces in the specification
        """
        return self.interfaces()

    def flattened(self):
        """Return an iterator of all included and extended interfaces
        """
        return iter(self.__iro__)

    def __sub__(self, other):
        """Remove interfaces from a specification
        """
        return Declaration(*[
            i for i in self.interfaces()
            if not [
                j
                for j in other.interfaces()
                if i.extends(j, 0) # non-strict extends
            ]
        ])

    def __add__(self, other):
        """
        Add two specifications or a specification and an interface
        and produce a new declaration.

        .. versionchanged:: 5.4.0
           Now tries to preserve a consistent resolution order. Interfaces
           being added to this object are added to the front of the resulting resolution
           order if they already extend an interface in this object. Previously,
           they were always added to the end of the order, which easily resulted in
           invalid orders.
        """
        before = []
        result = list(self.interfaces())
        seen = set(result)
        for i in other.interfaces():
            if i in seen:
                continue
            seen.add(i)
            if any(i.extends(x) for x in result):
                # It already extends us, e.g., is a subclass,
                # so it needs to go at the front of the RO.
                before.append(i)
            else:
                result.append(i)
        return Declaration(*(before + result))

    # XXX: Is __radd__ needed? No tests break if it's removed.
    # If it is needed, does it need to handle the C3 ordering differently?
    # I (JAM) don't *think* it does.
    __radd__ = __add__

    @staticmethod
    def _add_interfaces_to_cls(interfaces, cls):
        # Strip redundant interfaces already provided
        # by the cls so we don't produce invalid
        # resolution orders.
        implemented_by_cls = implementedBy(cls)
        interfaces = tuple([
            iface
            for iface in interfaces
            if not implemented_by_cls.isOrExtends(iface)
        ])
        return interfaces + (implemented_by_cls,)

    @staticmethod
    def _argument_names_for_repr(interfaces):
        # These don't actually have to be interfaces, they could be other
        # Specification objects like Implements. Also, the first
        # one is typically/nominally the cls.
        ordered_names = []
        names = set()
        for iface in interfaces:
            duplicate_transform = repr
            if isinstance(iface, InterfaceClass):
                # Special case to get 'foo.bar.IFace'
                # instead of '<InterfaceClass foo.bar.IFace>'
                this_name = iface.__name__
                duplicate_transform = str
            elif isinstance(iface, type):
                # Likewise for types. (Ignoring legacy old-style
                # classes.)
                this_name = iface.__name__
                duplicate_transform = _implements_name
            elif (isinstance(iface, Implements)
                  and not iface.declared
                  and iface.inherit in interfaces):
                # If nothing is declared, there's no need to even print this;
                # it would just show as ``classImplements(Class)``, and the
                # ``Class`` has typically already.
                continue
            else:
                this_name = repr(iface)

            already_seen = this_name in names
            names.add(this_name)
            if already_seen:
                this_name = duplicate_transform(iface)

            ordered_names.append(this_name)
        return ', '.join(ordered_names)


class _ImmutableDeclaration(Declaration):
    # A Declaration that is immutable. Used as a singleton to
    # return empty answers for things like ``implementedBy``.
    # We have to define the actual singleton after normalizeargs
    # is defined, and that in turn is defined after InterfaceClass and
    # Implements.

    __slots__ = ()

    __instance = None

    def __new__(cls):
        if _ImmutableDeclaration.__instance is None:
            _ImmutableDeclaration.__instance = object.__new__(cls)
        return _ImmutableDeclaration.__instance

    def __reduce__(self):
        return "_empty"

    @property
    def __bases__(self):
        return ()

    @__bases__.setter
    def __bases__(self, new_bases):
        # We expect the superclass constructor to set ``self.__bases__ = ()``.
        # Rather than attempt to special case that in the constructor and allow
        # setting __bases__ only at that time, it's easier to just allow setting
        # the empty tuple at any time. That makes ``x.__bases__ = x.__bases__`` a nice
        # no-op too. (Skipping the superclass constructor altogether is a recipe
        # for maintenance headaches.)
        if new_bases != ():
            raise TypeError("Cannot set non-empty bases on shared empty Declaration.")

    # As the immutable empty declaration, we cannot be changed.
    # This means there's no logical reason for us to have dependents
    # or subscriptions: we'll never notify them. So there's no need for
    # us to keep track of any of that.
    @property
    def dependents(self):
        return {}

    changed = subscribe = unsubscribe = lambda self, _ignored: None

    def interfaces(self):
        # An empty iterator
        return iter(())

    def extends(self, interface, strict=True):
        return interface is self._ROOT

    def get(self, name, default=None):
        return default

    def weakref(self, callback=None):
        # We're a singleton, we never go away. So there's no need to return
        # distinct weakref objects here; their callbacks will never
        # be called. Instead, we only need to return a callable that
        # returns ourself. The easiest one is to return _ImmutableDeclaration
        # itself; testing on Python 3.8 shows that's faster than a function that
        # returns _empty. (Remember, one goal is to avoid allocating any
        # object, and that includes a method.)
        return _ImmutableDeclaration

    @property
    def _v_attrs(self):
        # _v_attrs is not a public, documented property, but some client code
        # uses it anyway as a convenient place to cache things. To keep the
        # empty declaration truly immutable, we must ignore that. That includes
        # ignoring assignments as well.
        return {}

    @_v_attrs.setter
    def _v_attrs(self, new_attrs):
        pass


##############################################################################
#
# Implementation specifications
#
# These specify interfaces implemented by instances of classes

class Implements(NameAndModuleComparisonMixin,
                 Declaration):
    # Inherit from NameAndModuleComparisonMixin to be
    # mutually comparable with InterfaceClass objects.
    # (The two must be mutually comparable to be able to work in e.g., BTrees.)
    # Instances of this class generally don't have a __module__ other than
    # `zope.interface.declarations`, whereas they *do* have a __name__ that is the
    # fully qualified name of the object they are representing.

    # Note, though, that equality and hashing are still identity based. This
    # accounts for things like nested objects that have the same name (typically
    # only in tests) and is consistent with pickling. As far as comparisons to InterfaceClass
    # goes, we'll never have equal name and module to those, so we're still consistent there.
    # Instances of this class are essentially intended to be unique and are
    # heavily cached (note how our __reduce__ handles this) so having identity
    # based hash and eq should also work.

    # We want equality and hashing to be based on identity. However, we can't actually
    # implement __eq__/__ne__ to do this because sometimes we get wrapped in a proxy.
    # We need to let the proxy types implement these methods so they can handle unwrapping
    # and then rely on: (1) the interpreter automatically changing `implements == proxy` into
    # `proxy == implements` (which will call proxy.__eq__ to do the unwrapping) and then
    # (2) the default equality and hashing semantics being identity based.

    # class whose specification should be used as additional base
    inherit = None

    # interfaces actually declared for a class
    declared = ()

    # Weak cache of {class: <implements>} for super objects.
    # Created on demand. These are rare, as of 5.0 anyway. Using a class
    # level default doesn't take space in instances. Using _v_attrs would be
    # another place to store this without taking space unless needed.
    _super_cache = None

    __name__ = '?'

    @classmethod
    def named(cls, name, *bases):
        # Implementation method: Produce an Implements interface with
        # a fully fleshed out __name__ before calling the constructor, which
        # sets bases to the given interfaces and which may pass this object to
        # other objects (e.g., to adjust dependents). If they're sorting or comparing
        # by name, this needs to be set.
        inst = cls.__new__(cls)
        inst.__name__ = name
        inst.__init__(*bases)
        return inst

    def changed(self, originally_changed):
        try:
            del self._super_cache
        except AttributeError:
            pass
        return super().changed(originally_changed)

    def __repr__(self):
        if self.inherit:
            name = getattr(self.inherit, '__name__', None) or _implements_name(self.inherit)
        else:
            name = self.__name__
        declared_names = self._argument_names_for_repr(self.declared)
        if declared_names:
            declared_names = ', ' + declared_names
        return 'classImplements({}{})'.format(name, declared_names)

    def __reduce__(self):
        return implementedBy, (self.inherit, )


def _implements_name(ob):
    # Return the __name__ attribute to be used by its __implemented__
    # property.
    # This must be stable for the "same" object across processes
    # because it is used for sorting. It needn't be unique, though, in cases
    # like nested classes named Foo created by different functions, because
    # equality and hashing is still based on identity.
    # It might be nice to use __qualname__ on Python 3, but that would produce
    # different values between Py2 and Py3.
    return (getattr(ob, '__module__', '?') or '?') + \
        '.' + (getattr(ob, '__name__', '?') or '?')


def _implementedBy_super(sup):
    # TODO: This is now simple enough we could probably implement
    # in C if needed.

    # If the class MRO is strictly linear, we could just
    # follow the normal algorithm for the next class in the
    # search order (e.g., just return
    # ``implemented_by_next``). But when diamond inheritance
    # or mixins + interface declarations are present, we have
    # to consider the whole MRO and compute a new Implements
    # that excludes the classes being skipped over but
    # includes everything else.
    implemented_by_self = implementedBy(sup.__self_class__)
    cache = implemented_by_self._super_cache # pylint:disable=protected-access
    if cache is None:
        cache = implemented_by_self._super_cache = weakref.WeakKeyDictionary()

    key = sup.__thisclass__
    try:
        return cache[key]
    except KeyError:
        pass

    next_cls = _next_super_class(sup)
    # For ``implementedBy(cls)``:
    # .__bases__ is .declared + [implementedBy(b) for b in cls.__bases__]
    # .inherit is cls

    implemented_by_next = implementedBy(next_cls)
    mro = sup.__self_class__.__mro__
    ix_next_cls = mro.index(next_cls)
    classes_to_keep = mro[ix_next_cls:]
    new_bases = [implementedBy(c) for c in classes_to_keep]

    new = Implements.named(
        implemented_by_self.__name__ + ':' + implemented_by_next.__name__,
        *new_bases
    )
    new.inherit = implemented_by_next.inherit
    new.declared = implemented_by_next.declared
    # I don't *think* that new needs to subscribe to ``implemented_by_self``;
    # it auto-subscribed to its bases, and that should be good enough.
    cache[key] = new

    return new


@_use_c_impl
def implementedBy(cls): # pylint:disable=too-many-return-statements,too-many-branches
    """Return the interfaces implemented for a class' instances

      The value returned is an `~zope.interface.interfaces.IDeclaration`.
    """
    try:
        if isinstance(cls, super):
            # Yes, this needs to be inside the try: block. Some objects
            # like security proxies even break isinstance.
            return _implementedBy_super(cls)

        spec = cls.__dict__.get('__implemented__')
    except AttributeError:

        # we can't get the class dict. This is probably due to a
        # security proxy.  If this is the case, then probably no
        # descriptor was installed for the class.

        # We don't want to depend directly on zope.security in
        # zope.interface, but we'll try to make reasonable
        # accommodations in an indirect way.

        # We'll check to see if there's an implements:

        spec = getattr(cls, '__implemented__', None)
        if spec is None:
            # There's no spec stred in the class. Maybe its a builtin:
            spec = BuiltinImplementationSpecifications.get(cls)
            if spec is not None:
                return spec
            return _empty

        if spec.__class__ == Implements:
            # we defaulted to _empty or there was a spec. Good enough.
            # Return it.
            return spec

        # TODO: need old style __implements__ compatibility?
        # Hm, there's an __implemented__, but it's not a spec. Must be
        # an old-style declaration. Just compute a spec for it
        return Declaration(*_normalizeargs((spec, )))

    if isinstance(spec, Implements):
        return spec

    if spec is None:
        spec = BuiltinImplementationSpecifications.get(cls)
        if spec is not None:
            return spec

    # TODO: need old style __implements__ compatibility?
    spec_name = _implements_name(cls)
    if spec is not None:
        # old-style __implemented__ = foo declaration
        spec = (spec, ) # tuplefy, as it might be just an int
        spec = Implements.named(spec_name, *_normalizeargs(spec))
        spec.inherit = None    # old-style implies no inherit
        del cls.__implemented__ # get rid of the old-style declaration
    else:
        try:
            bases = cls.__bases__
        except AttributeError:
            if not callable(cls):
                raise TypeError("ImplementedBy called for non-factory", cls)
            bases = ()

        spec = Implements.named(spec_name, *[implementedBy(c) for c in bases])
        spec.inherit = cls

    try:
        cls.__implemented__ = spec
        if not hasattr(cls, '__providedBy__'):
            cls.__providedBy__ = objectSpecificationDescriptor

        if isinstance(cls, type) and '__provides__' not in cls.__dict__:
            # Make sure we get a __provides__ descriptor
            cls.__provides__ = ClassProvides(
                cls,
                getattr(cls, '__class__', type(cls)),
                )

    except TypeError:
        if not isinstance(cls, type):
            raise TypeError("ImplementedBy called for non-type", cls)
        BuiltinImplementationSpecifications[cls] = spec

    return spec


def classImplementsOnly(cls, *interfaces):
    """
    Declare the only interfaces implemented by instances of a class

    The arguments after the class are one or more interfaces or interface
    specifications (`~zope.interface.interfaces.IDeclaration` objects).

    The interfaces given (including the interfaces in the specifications)
    replace any previous declarations, *including* inherited definitions. If you
    wish to preserve inherited declarations, you can pass ``implementedBy(cls)``
    in *interfaces*. This can be used to alter the interface resolution order.
    """
    spec = implementedBy(cls)
    # Clear out everything inherited. It's important to
    # also clear the bases right now so that we don't improperly discard
    # interfaces that are already implemented by *old* bases that we're
    # about to get rid of.
    spec.declared = ()
    spec.inherit = None
    spec.__bases__ = ()
    _classImplements_ordered(spec, interfaces, ())


def classImplements(cls, *interfaces):
    """
    Declare additional interfaces implemented for instances of a class

    The arguments after the class are one or more interfaces or
    interface specifications (`~zope.interface.interfaces.IDeclaration` objects).

    The interfaces given (including the interfaces in the specifications)
    are added to any interfaces previously declared. An effort is made to
    keep a consistent C3 resolution order, but this cannot be guaranteed.

    .. versionchanged:: 5.0.0
       Each individual interface in *interfaces* may be added to either the
       beginning or end of the list of interfaces declared for *cls*,
       based on inheritance, in order to try to maintain a consistent
       resolution order. Previously, all interfaces were added to the end.
    .. versionchanged:: 5.1.0
       If *cls* is already declared to implement an interface (or derived interface)
       in *interfaces* through inheritance, the interface is ignored. Previously, it
       would redundantly be made direct base of *cls*, which often produced inconsistent
       interface resolution orders. Now, the order will be consistent, but may change.
       Also, if the ``__bases__`` of the *cls* are later changed, the *cls* will no
       longer be considered to implement such an interface (changing the ``__bases__`` of *cls*
       has never been supported).
    """
    spec = implementedBy(cls)
    interfaces = tuple(_normalizeargs(interfaces))

    before = []
    after = []

    # Take steps to try to avoid producing an invalid resolution
    # order, while still allowing for BWC (in the past, we always
    # appended)
    for iface in interfaces:
        for b in spec.declared:
            if iface.extends(b):
                before.append(iface)
                break
        else:
            after.append(iface)
    _classImplements_ordered(spec, tuple(before), tuple(after))


def classImplementsFirst(cls, iface):
    """
    Declare that instances of *cls* additionally provide *iface*.

    The second argument is an interface or interface specification.
    It is added as the highest priority (first in the IRO) interface;
    no attempt is made to keep a consistent resolution order.

    .. versionadded:: 5.0.0
    """
    spec = implementedBy(cls)
    _classImplements_ordered(spec, (iface,), ())


def _classImplements_ordered(spec, before=(), after=()):
    # Elide everything already inherited.
    # Except, if it is the root, and we don't already declare anything else
    # that would imply it, allow the root through. (TODO: When we disallow non-strict
    # IRO, this part of the check can be removed because it's not possible to re-declare
    # like that.)
    before = [
        x
        for x in before
        if not spec.isOrExtends(x) or (x is Interface and not spec.declared)
    ]
    after = [
        x
        for x in after
        if not spec.isOrExtends(x) or (x is Interface and not spec.declared)
    ]

    # eliminate duplicates
    new_declared = []
    seen = set()
    for l in before, spec.declared, after:
        for b in l:
            if b not in seen:
                new_declared.append(b)
                seen.add(b)

    spec.declared = tuple(new_declared)

    # compute the bases
    bases = new_declared # guaranteed no dupes

    if spec.inherit is not None:
        for c in spec.inherit.__bases__:
            b = implementedBy(c)
            if b not in seen:
                seen.add(b)
                bases.append(b)

    spec.__bases__ = tuple(bases)


def _implements_advice(cls):
    interfaces, do_classImplements = cls.__dict__['__implements_advice_data__']
    del cls.__implements_advice_data__
    do_classImplements(cls, *interfaces)
    return cls


class implementer:
    """
    Declare the interfaces implemented by instances of a class.

    This function is called as a class decorator.

    The arguments are one or more interfaces or interface
    specifications (`~zope.interface.interfaces.IDeclaration`
    objects).

    The interfaces given (including the interfaces in the
    specifications) are added to any interfaces previously declared,
    unless the interface is already implemented.

    Previous declarations include declarations for base classes unless
    implementsOnly was used.

    This function is provided for convenience. It provides a more
    convenient way to call `classImplements`. For example::

        @implementer(I1)
        class C(object):
            pass

    is equivalent to calling::

        classImplements(C, I1)

    after the class has been created.

    .. seealso:: `classImplements`
       The change history provided there applies to this function too.
    """
    __slots__ = ('interfaces',)

    def __init__(self, *interfaces):
        self.interfaces = interfaces

    def __call__(self, ob):
        if isinstance(ob, type):
            # This is the common branch for classes.
            classImplements(ob, *self.interfaces)
            return ob

        spec_name = _implements_name(ob)
        spec = Implements.named(spec_name, *self.interfaces)
        try:
            ob.__implemented__ = spec
        except AttributeError:
            raise TypeError("Can't declare implements", ob)
        return ob

class implementer_only:
    """Declare the only interfaces implemented by instances of a class

      This function is called as a class decorator.

      The arguments are one or more interfaces or interface
      specifications (`~zope.interface.interfaces.IDeclaration` objects).

      Previous declarations including declarations for base classes
      are overridden.

      This function is provided for convenience. It provides a more
      convenient way to call `classImplementsOnly`. For example::

        @implementer_only(I1)
        class C(object): pass

      is equivalent to calling::

        classImplementsOnly(I1)

      after the class has been created.
      """

    def __init__(self, *interfaces):
        self.interfaces = interfaces

    def __call__(self, ob):
        if isinstance(ob, (FunctionType, MethodType)):
            # XXX Does this decorator make sense for anything but classes?
            # I don't think so. There can be no inheritance of interfaces
            # on a method or function....
            raise ValueError('The implementer_only decorator is not '
                             'supported for methods or functions.')

        # Assume it's a class:
        classImplementsOnly(ob, *self.interfaces)
        return ob


##############################################################################
#
# Instance declarations

class Provides(Declaration):  # Really named ProvidesClass
    """Implement ``__provides__``, the instance-specific specification

    When an object is pickled, we pickle the interfaces that it implements.
    """

    def __init__(self, cls, *interfaces):
        self.__args = (cls, ) + interfaces
        self._cls = cls
        Declaration.__init__(self, *self._add_interfaces_to_cls(interfaces, cls))

    # Added to by ``moduleProvides``, et al
    _v_module_names = ()

    def __repr__(self):
        # The typical way to create instances of this
        # object is via calling ``directlyProvides(...)`` or ``alsoProvides()``,
        # but that's not the only way. Proxies, for example,
        # directly use the ``Provides(...)`` function (which is the
        # more generic method, and what we pickle as). We're after the most
        # readable, useful repr in the common case, so we use the most
        # common name.
        #
        # We also cooperate with ``moduleProvides`` to attempt to do the
        # right thing for that API. See it for details.
        function_name = 'directlyProvides'
        if self._cls is ModuleType and self._v_module_names:
            # See notes in ``moduleProvides``/``directlyProvides``
            providing_on_module = True
            interfaces = self.__args[1:]
        else:
            providing_on_module = False
            interfaces = (self._cls,) + self.__bases__
        ordered_names = self._argument_names_for_repr(interfaces)
        if providing_on_module:
            mod_names = self._v_module_names
            if len(mod_names) == 1:
                mod_names = "sys.modules[%r]" % mod_names[0]
            ordered_names = (
                '{}, '.format(mod_names)
            ) + ordered_names
        return "{}({})".format(
            function_name,
            ordered_names,
        )

    def __reduce__(self):
        # This reduces to the Provides *function*, not
        # this class.
        return Provides, self.__args

    __module__ = 'zope.interface'

    def __get__(self, inst, cls):
        """Make sure that a class __provides__ doesn't leak to an instance
        """
        if inst is None and cls is self._cls:
            # We were accessed through a class, so we are the class'
            # provides spec. Just return this object, but only if we are
            # being called on the same class that we were defined for:
            return self

        raise AttributeError('__provides__')

ProvidesClass = Provides

# Registry of instance declarations
# This is a memory optimization to allow objects to share specifications.
InstanceDeclarations = weakref.WeakValueDictionary()

def Provides(*interfaces): # pylint:disable=function-redefined
    """Cache instance declarations

      Instance declarations are shared among instances that have the same
      declaration. The declarations are cached in a weak value dictionary.
    """
    spec = InstanceDeclarations.get(interfaces)
    if spec is None:
        spec = ProvidesClass(*interfaces)
        InstanceDeclarations[interfaces] = spec

    return spec

Provides.__safe_for_unpickling__ = True


def directlyProvides(object, *interfaces): # pylint:disable=redefined-builtin
    """Declare interfaces declared directly for an object

      The arguments after the object are one or more interfaces or interface
      specifications (`~zope.interface.interfaces.IDeclaration` objects).

      The interfaces given (including the interfaces in the specifications)
      replace interfaces previously declared for the object.
    """
    cls = getattr(object, '__class__', None)
    if cls is not None and getattr(cls, '__class__', None) is cls:
        # It's a meta class (well, at least it it could be an extension class)
        # Note that we can't get here from the tests:  there is no normal
        # class which isn't descriptor aware.
        if not isinstance(object, type):
            raise TypeError("Attempt to make an interface declaration on a "
                            "non-descriptor-aware class")

    interfaces = _normalizeargs(interfaces)
    if cls is None:
        cls = type(object)

    if issubclass(cls, type):
        # we have a class or type.  We'll use a special descriptor
        # that provides some extra caching
        object.__provides__ = ClassProvides(object, cls, *interfaces)
    else:
        provides = object.__provides__ = Provides(cls, *interfaces)
        # See notes in ``moduleProvides``.
        if issubclass(cls, ModuleType) and hasattr(object, '__name__'):
            provides._v_module_names += (object.__name__,)



def alsoProvides(object, *interfaces): # pylint:disable=redefined-builtin
    """Declare interfaces declared directly for an object

    The arguments after the object are one or more interfaces or interface
    specifications (`~zope.interface.interfaces.IDeclaration` objects).

    The interfaces given (including the interfaces in the specifications) are
    added to the interfaces previously declared for the object.
    """
    directlyProvides(object, directlyProvidedBy(object), *interfaces)


def noLongerProvides(object, interface): # pylint:disable=redefined-builtin
    """ Removes a directly provided interface from an object.
    """
    directlyProvides(object, directlyProvidedBy(object) - interface)
    if interface.providedBy(object):
        raise ValueError("Can only remove directly provided interfaces.")


@_use_c_impl
class ClassProvidesBase(SpecificationBase):

    __slots__ = (
        '_cls',
        '_implements',
    )

    def __get__(self, inst, cls):
        # member slots are set by subclass
        # pylint:disable=no-member
        if cls is self._cls:
            # We only work if called on the class we were defined for

            if inst is None:
                # We were accessed through a class, so we are the class'
                # provides spec. Just return this object as is:
                return self

            return self._implements

        raise AttributeError('__provides__')


class ClassProvides(Declaration, ClassProvidesBase):
    """Special descriptor for class ``__provides__``

    The descriptor caches the implementedBy info, so that
    we can get declarations for objects without instance-specific
    interfaces a bit quicker.
    """

    __slots__ = (
        '__args',
    )

    def __init__(self, cls, metacls, *interfaces):
        self._cls = cls
        self._implements = implementedBy(cls)
        self.__args = (cls, metacls, ) + interfaces
        Declaration.__init__(self, *self._add_interfaces_to_cls(interfaces, metacls))

    def __repr__(self):
        # There are two common ways to get instances of this object:
        # The most interesting way is calling ``@provider(..)`` as a decorator
        # of a class; this is the same as calling ``directlyProvides(cls, ...)``.
        #
        # The other way is by default: anything that invokes ``implementedBy(x)``
        # will wind up putting an instance in ``type(x).__provides__``; this includes
        # the ``@implementer(...)`` decorator. Those instances won't have any
        # interfaces.
        #
        # Thus, as our repr, we go with the ``directlyProvides()`` syntax.
        interfaces = (self._cls, ) + self.__args[2:]
        ordered_names = self._argument_names_for_repr(interfaces)
        return "directlyProvides({})".format(ordered_names)

    def __reduce__(self):
        return self.__class__, self.__args

    # Copy base-class method for speed
    __get__ = ClassProvidesBase.__get__


def directlyProvidedBy(object): # pylint:disable=redefined-builtin
    """Return the interfaces directly provided by the given object

    The value returned is an `~zope.interface.interfaces.IDeclaration`.
    """
    provides = getattr(object, "__provides__", None)
    if (
            provides is None # no spec
            # We might have gotten the implements spec, as an
            # optimization. If so, it's like having only one base, that we
            # lop off to exclude class-supplied declarations:
            or isinstance(provides, Implements)
    ):
        return _empty

    # Strip off the class part of the spec:
    return Declaration(provides.__bases__[:-1])


class provider:
    """Declare interfaces provided directly by a class

      This function is called in a class definition.

      The arguments are one or more interfaces or interface specifications
      (`~zope.interface.interfaces.IDeclaration` objects).

      The given interfaces (including the interfaces in the specifications)
      are used to create the class's direct-object interface specification.
      An error will be raised if the module class has an direct interface
      specification. In other words, it is an error to call this function more
      than once in a class definition.

      Note that the given interfaces have nothing to do with the interfaces
      implemented by instances of the class.

      This function is provided for convenience. It provides a more convenient
      way to call `directlyProvides` for a class. For example::

        @provider(I1)
        class C:
            pass

      is equivalent to calling::

        directlyProvides(C, I1)

      after the class has been created.
    """

    def __init__(self, *interfaces):
        self.interfaces = interfaces

    def __call__(self, ob):
        directlyProvides(ob, *self.interfaces)
        return ob


def moduleProvides(*interfaces):
    """Declare interfaces provided by a module

    This function is used in a module definition.

    The arguments are one or more interfaces or interface specifications
    (`~zope.interface.interfaces.IDeclaration` objects).

    The given interfaces (including the interfaces in the specifications) are
    used to create the module's direct-object interface specification.  An
    error will be raised if the module already has an interface specification.
    In other words, it is an error to call this function more than once in a
    module definition.

    This function is provided for convenience. It provides a more convenient
    way to call directlyProvides. For example::

      moduleProvides(I1)

    is equivalent to::

      directlyProvides(sys.modules[__name__], I1)
    """
    frame = sys._getframe(1) # pylint:disable=protected-access
    locals = frame.f_locals # pylint:disable=redefined-builtin

    # Try to make sure we were called from a module body
    if (locals is not frame.f_globals) or ('__name__' not in locals):
        raise TypeError(
            "moduleProvides can only be used from a module definition.")

    if '__provides__' in locals:
        raise TypeError(
            "moduleProvides can only be used once in a module definition.")

    # Note: This is cached based on the key ``(ModuleType, *interfaces)``;
    # One consequence is that any module that provides the same interfaces
    # gets the same ``__repr__``, meaning that you can't tell what module
    # such a declaration came from. Adding the module name to ``_v_module_names``
    # attempts to correct for this; it works in some common situations, but fails
    # (1) after pickling (the data is lost) and (2) if declarations are
    # actually shared and (3) if the alternate spelling of ``directlyProvides()``
    # is used. Problem (3)  is fixed by cooperating with ``directlyProvides``
    # to maintain this information, and problem (2) is worked around by
    # printing all the names, but (1) is unsolvable without introducing
    # new classes or changing the stored data...but it doesn't actually matter,
    # because ``ModuleType`` can't be pickled!
    p = locals["__provides__"] = Provides(ModuleType,
                                          *_normalizeargs(interfaces))
    p._v_module_names += (locals['__name__'],)


##############################################################################
#
# Declaration querying support

# XXX:  is this a fossil?  Nobody calls it, no unit tests exercise it, no
#       doctests import it, and the package __init__ doesn't import it.
#       (Answer: Versions of zope.container prior to 4.4.0 called this,
#        and zope.proxy.decorator up through at least 4.3.5 called this.)
def ObjectSpecification(direct, cls):
    """Provide object specifications

    These combine information for the object and for it's classes.
    """
    return Provides(cls, direct) # pragma: no cover fossil

@_use_c_impl
def getObjectSpecification(ob):
    try:
        provides = ob.__provides__
    except AttributeError:
        provides = None

    if provides is not None:
        if isinstance(provides, SpecificationBase):
            return provides

    try:
        cls = ob.__class__
    except AttributeError:
        # We can't get the class, so just consider provides
        return _empty
    return implementedBy(cls)


@_use_c_impl
def providedBy(ob):
    """
    Return the interfaces provided by *ob*.

    If *ob* is a :class:`super` object, then only interfaces implemented
    by the remainder of the classes in the method resolution order are
    considered. Interfaces directly provided by the object underlying *ob*
    are not.
    """
    # Here we have either a special object, an old-style declaration
    # or a descriptor

    # Try to get __providedBy__
    try:
        if isinstance(ob, super): # Some objects raise errors on isinstance()
            return implementedBy(ob)

        r = ob.__providedBy__
    except AttributeError:
        # Not set yet. Fall back to lower-level thing that computes it
        return getObjectSpecification(ob)

    try:
        # We might have gotten a descriptor from an instance of a
        # class (like an ExtensionClass) that doesn't support
        # descriptors.  We'll make sure we got one by trying to get
        # the only attribute, which all specs have.
        r.extends
    except AttributeError:

        # The object's class doesn't understand descriptors.
        # Sigh. We need to get an object descriptor, but we have to be
        # careful.  We want to use the instance's __provides__, if
        # there is one, but only if it didn't come from the class.

        try:
            r = ob.__provides__
        except AttributeError:
            # No __provides__, so just fall back to implementedBy
            return implementedBy(ob.__class__)

        # We need to make sure we got the __provides__ from the
        # instance. We'll do this by making sure we don't get the same
        # thing from the class:

        try:
            cp = ob.__class__.__provides__
        except AttributeError:
            # The ob doesn't have a class or the class has no
            # provides, assume we're done:
            return r

        if r is cp:
            # Oops, we got the provides from the class. This means
            # the object doesn't have it's own. We should use implementedBy
            return implementedBy(ob.__class__)

    return r


@_use_c_impl
class ObjectSpecificationDescriptor:
    """Implement the ``__providedBy__`` attribute

    The ``__providedBy__`` attribute computes the interfaces provided by
    an object. If an object has an ``__provides__`` attribute, that is returned.
    Otherwise, `implementedBy` the *cls* is returned.

    .. versionchanged:: 5.4.0
       Both the default (C) implementation and the Python implementation
       now let exceptions raised by accessing ``__provides__`` propagate.
       Previously, the C version ignored all exceptions.
    .. versionchanged:: 5.4.0
       The Python implementation now matches the C implementation and lets
       a ``__provides__`` of ``None`` override what the class is declared to
       implement.
    """

    def __get__(self, inst, cls):
        """Get an object specification for an object
        """
        if inst is None:
            return getObjectSpecification(cls)

        try:
            return inst.__provides__
        except AttributeError:
            return implementedBy(cls)


##############################################################################

def _normalizeargs(sequence, output=None):
    """Normalize declaration arguments

    Normalization arguments might contain Declarions, tuples, or single
    interfaces.

    Anything but individual interfaces or implements specs will be expanded.
    """
    if output is None:
        output = []

    cls = sequence.__class__
    if InterfaceClass in cls.__mro__ or Implements in cls.__mro__:
        output.append(sequence)
    else:
        for v in sequence:
            _normalizeargs(v, output)

    return output

_empty = _ImmutableDeclaration()

objectSpecificationDescriptor = ObjectSpecificationDescriptor()
